# -*- coding: utf-8 -*-
"""Functions that operate on lists, dicts, and other objects.

.. versionadded:: 1.0.0
"""

from __future__ import absolute_import

import copy
from functools import partial
import math
import re

import pydash as pyd
from .helpers import (
    NoValue,
    callit,
    iterator,
    iteriteratee,
    getargcount,
    base_get,
    base_set
)
from ._compat import iteritems, string_types, text_type
from .utilities import PathToken, to_path, to_path_tokens


__all__ = (
    'assign',
    'assign_with',
    'callables',
    'clone',
    'clone_deep',
    'clone_deep_with',
    'clone_with',
    'defaults',
    'defaults_deep',
    'find_key',
    'find_last_key',
    'for_in',
    'for_in_right',
    'get',
    'has',
    'invert',
    'invert_by',
    'invoke',
    'keys',
    'map_keys',
    'map_values',
    'map_values_deep',
    'merge',
    'merge_with',
    'omit',
    'omit_by',
    'parse_int',
    'pick',
    'pick_by',
    'rename_keys',
    'set_',
    'set_with',
    'to_boolean',
    'to_dict',
    'to_integer',
    'to_list',
    'to_number',
    'to_pairs',
    'to_string',
    'transform',
    'unset',
    'update',
    'update_with',
    'values',
)


def assign(obj, *sources):
    """Assigns properties of source object(s) to the destination object.

    Args:
        obj (dict): Destination object whose properties will be modified.
        sources (dict): Source objects to assign to `obj`.

    Returns:
        dict: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> obj = {}
        >>> obj2 = assign(obj, {'a': 1}, {'b': 2}, {'c': 3})
        >>> obj == {'a': 1, 'b': 2, 'c': 3}
        True
        >>> obj is obj2
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.3.2
        Apply :func:`clone_deep` to each `source` before assigning to `obj`.

    .. versionchanged:: 3.0.0
        Allow iteratees to accept partial arguments.

    .. versionchanged:: 3.4.4
        Shallow copy each `source` instead of deep copying.

    .. versionchanged:: 4.0.0

        - Moved `iteratee` argument to :func:`assign_with`.
        - Removed alias ``extend``.
    """
    return assign_with(obj, *sources)


def assign_with(obj, *sources, **kargs):
    """This method is like :func:`assign` except that it accepts customizer
    which is invoked to produce the assigned values. If customizer returns
    ``None``, assignment is handled by the method instead. The customizer is
    invoked with five arguments: ``(obj_value, src_value, key, obj, source)``.

    Args:
        obj (dict): Destination object whose properties will be modified.
        sources (dict): Source objects to assign to `obj`.

    Keyword Args:
        customizer (mixed, optional): Customizer applied per iteration.

    Returns:
        dict: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> customizer = lambda o, s: s if o is None else o
        >>> results = assign({'a': 1}, {'b': 2}, {'a': 3}, customizer)
        >>> results == {'a': 1, 'b': 2}
        True

    .. versionadded:: 4.0.0
    """
    sources = list(sources)
    customizer = kargs.get('customizer')

    if customizer is None and callable(sources[-1]):
        customizer = sources.pop()

    if customizer is not None:
        argcount = getargcount(customizer, maxargs=5)
    else:
        argcount = None

    for source in sources:
        source = source.copy()

        for key, value in iteritems(source):
            if customizer:
                val = callit(customizer,
                             obj.get(key),
                             value,
                             key,
                             obj,
                             source,
                             argcount=argcount)
                if val is not None:
                    value = val

            obj[key] = value

    return obj


def callables(obj):
    """Creates a sorted list of keys of an object that are callable.

    Args:
        obj (list|dict): Object to inspect.

    Returns:
        list: All keys whose values are callable.

    Example:

        >>> callables({'a': 1, 'b': lambda: 2, 'c': lambda: 3})
        ['b', 'c']

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0
        Renamed ``functions`` to ``callables``.

    .. versionchanged:: 4.0.0
        Removed alias ``methods``.
    """
    return sorted(key for key, value in iterator(obj) if callable(value))


def clone(value):
    """Creates a clone of `value`.

    Args:
        value (list|dict): Object to clone.

    Example:

        >>> x = {'a': 1, 'b': 2, 'c': {'d': 3}}
        >>> y = clone(x)
        >>> y == y
        True
        >>> x is y
        False
        >>> x['c'] is y['c']
        True

    Returns:
        list|dict: Cloned object.

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved 'iteratee' parameter to :func:`clone_with`.
    """
    return base_clone(value)


def clone_with(value, customizer=None):
    """This method is like :func:`clone` except that it accepts customizer
    which is invoked to produce the cloned value. If customizer returns
    ``None``, cloning is handled by the method instead. The customizer is
    invoked with up to three arguments: ``(value, index|key, object)``.

    Args:
        value (list|dict): Object to clone.
        customizer (callable, optional): Function to customize cloning.

    Returns:
        list|dict: Cloned object.

    Example:

        >>> x = {'a': 1, 'b': 2, 'c': {'d': 3}}
        >>> cbk = lambda v, k: v + 2 if isinstance(v, int) and k else None
        >>> y = clone_with(x, cbk)
        >>> y == {'a': 3, 'b': 4, 'c': {'d': 3}}
        True
    """
    return base_clone(value, customizer=customizer)


def clone_deep(value):
    """Creates a deep clone of `value`. If a iteratee is provided it will be
    executed to produce the cloned values.

    Args:
        value (list|dict): Object to clone.

    Returns:
        list|dict: Cloned object.

    Example:

        >>> x = {'a': 1, 'b': 2, 'c': {'d': 3}}
        >>> y = clone_deep(x)
        >>> y == y
        True
        >>> x is y
        False
        >>> x['c'] is y['c']
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved 'iteratee' parameter to :func:`clone_deep_with`.
    """
    return base_clone(value, is_deep=True)


def clone_deep_with(value, customizer=None):
    """This method is like :func:`clone_with` except that it recursively clones
    `value`.

    Args:
        value (list|dict): Object to clone.
        customizer (callable, optional): Function to customize cloning.

    Returns:
        list|dict: Cloned object.
    """
    return base_clone(value, is_deep=True, customizer=customizer)


def defaults(obj, *sources):
    """Assigns properties of source object(s) to the destination object for all
    destination properties that resolve to undefined.

    Args:
        obj (dict): Destination object whose properties will be modified.
        sources (dict): Source objects to assign to `obj`.

    Returns:
        dict: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> obj = {'a': 1}
        >>> obj2 = defaults(obj, {'b': 2}, {'c': 3}, {'a': 4})
        >>> obj is obj2
        True
        >>> obj == {'a': 1, 'b': 2, 'c': 3}
        True

    .. versionadded:: 1.0.0
    """
    for source in sources:
        for key, value in iteritems(source):
            obj.setdefault(key, value)

    return obj


def defaults_deep(obj, *sources):
    """This method is like :func:`defaults` except that it recursively assigns
    default properties.

    Args:
        obj (dict): Destination object whose properties will be modified.
        sources (dict): Source objects to assign to `obj`.

    Returns:
        dict: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> obj = {'a': {'b': 1}}
        >>> obj2 = defaults_deep(obj, {'a': {'b': 2, 'c': 3}})
        >>> obj is obj2
        True
        >>> obj == {'a': {'b': 1, 'c': 3}}
        True

    .. versionadded:: 3.3.0
    """
    def setter(obj, key, value):
        if hasattr(obj, 'setdefault'):
            obj.setdefault(key, value)

    return merge_with(obj, *sources, _setter=setter)


def find_key(obj, predicate=None):
    """This method is like :func:`pydash.arrays.find_index` except that it
    returns the key of the first element that passes the predicate check,
    instead of the element itself.

    Args:
        obj (list|dict): Object to search.
        predicate (mixed): Predicate applied per iteration.

    Returns:
        mixed: Found key or ``None``.

    Example:

        >>> find_key({'a': 1, 'b': 2, 'c': 3}, lambda x: x == 1)
        'a'
        >>> find_key([1, 2, 3, 4], lambda x: x == 1)
        0

    .. versionadded:: 1.0.0
    """
    for result, _, key, _ in iteriteratee(obj, predicate):
        if result:
            return key


def find_last_key(obj, predicate=None):
    """This method is like :func:`find_key` except that it iterates over
    elements of a collection in the opposite order.

    Args:
        obj (list|dict): Object to search.
        predicate (mixed): Predicate applied per iteration.

    Returns:
        mixed: Found key or ``None``.

    Example:

        >>> find_last_key({'a': 1, 'b': 2, 'c': 3}, lambda x: x == 1)
        'a'
        >>> find_last_key([1, 2, 3, 1], lambda x: x == 1)
        3

    .. versionchanged:: 4.0.0
        Made into its own function (instead of an alias of ``find_key``) with
        proper reverse find implementation.
    """
    reversed_obj = reversed(list(iteriteratee(obj, predicate)))
    for result, _, key, _ in reversed_obj:
        if result:
            return key


def for_in(obj, iteratee=None):
    """Iterates over own and inherited enumerable properties of `obj`,
    executing `iteratee` for each property.

    Args:
        obj (list|dict): Object to process.
        iteratee (mixed): Iteratee applied per iteration.

    Returns:
        list|dict: `obj`.

    Example:

        >>> obj = {}
        >>> def cb(v, k): obj[k] = v
        >>> results = for_in({'a': 1, 'b': 2, 'c': 3}, cb)
        >>> results == {'a': 1, 'b': 2, 'c': 3}
        True
        >>> obj == {'a': 1, 'b': 2, 'c': 3}
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``for_own``.
    """
    walk = (None for ret, _, _, _ in iteriteratee(obj, iteratee)
            if ret is False)
    next(walk, None)
    return obj


def for_in_right(obj, iteratee=None):
    """This function is like :func:`for_in` except it iterates over the
    properties in reverse order.

    Args:
        obj (list|dict): Object to process.
        iteratee (mixed): Iteratee applied per iteration.

    Returns:
        list|dict: `obj`.

    Example:

        >>> data = {'product': 1}
        >>> def cb(v): data['product'] *= v
        >>> for_in_right([1, 2, 3, 4], cb)
        [1, 2, 3, 4]
        >>> data['product'] == 24
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``for_own_right``.
    """
    walk = (None for ret, _, _, _ in iteriteratee(obj, iteratee, reverse=True)
            if ret is False)
    next(walk, None)
    return obj


def get(obj, path, default=None):
    """Get the value at any depth of a nested object based on the path
    described by `path`. If path doesn't exist, `default` is returned.

    Args:
        obj (list|dict): Object to process.
        path (str|list): List or ``.`` delimited string of path describing
            path.

    Keyword Arguments:
        default (mixed): Default value to return if path doesn't exist.
            Defaults to ``None``.

    Returns:
        mixed: Value of `obj` at path.

    Example:

        >>> get({}, 'a.b.c') is None
        True
        >>> get({'a': {'b': {'c': [1, 2, 3, 4]}}}, 'a.b.c[1]')
        2
        >>> get({'a': {'b': {'c': [1, 2, 3, 4]}}}, 'a.b.c.1')
        2
        >>> get({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.1')
        2
        >>> get({'a': {'b': [0, {'c': [1, 2]}]}}, ['a', 'b', 1, 'c', 1])
        2
        >>> get({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.2') is None
        True

    .. versionadded:: 2.0.0

    .. versionchanged:: 2.2.0
        Support escaping "." delimiter in single string path key.

    .. versionchanged:: 3.3.0

        - Added :func:`get` as main definition and :func:`get_path` as alias.
        - Made :func:`deep_get` an alias.

    .. versionchanged:: 3.4.7
        Fixed bug where an iterable default was iterated over instead of being
        returned when an object path wasn't found.

    .. versionchanged:: 4.0.0

        - Support attribute access on `obj` if item access fails.
        - Removed aliases ``get_path`` and ``deep_get``.
    """
    if default is NoValue:
        # When NoValue given for default, then this method will raise if path
        # is not present in obj.
        sentinel = default
    else:
        # When a returnable default is given, use a sentinel value to detect
        # when base_get() returns a default value for a missing path so we can
        # exit early from the loop and not mistakenly iterate over the default.
        sentinel = object()

    for key in to_path(path):
        obj = base_get(obj, key, default=sentinel)

        if obj is sentinel:
            # Path doesn't exist so set return obj to the default.
            obj = default
            break

    return obj


def has(obj, path):
    """Checks if `path` exists as a key of `obj`.

    Args:
        obj (mixed): Object to test.
        path (mixed): Path to test for. Can be a list of nested keys or a ``.``
            delimited string of path describing the path.

    Returns:
        bool: Whether `obj` has `path`.

    Example:

        >>> has([1, 2, 3], 1)
        True
        >>> has({'a': 1, 'b': 2}, 'b')
        True
        >>> has({'a': 1, 'b': 2}, 'c')
        False
        >>> has({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.1')
        True
        >>> has({'a': {'b': [0, {'c': [1, 2]}]}}, 'a.b.1.c.2')
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Return ``False`` on ``ValueError`` when checking path.

    .. versionchanged:: 3.3.0

        - Added :func:`deep_has` as alias.
        - Added :func:`has_path` as alias.

    .. versionchanged:: 4.0.0
        Removed aliases ``deep_has`` and ``has_path``.
    """
    try:
        get(obj, path, default=NoValue)
        exists = True
    except (KeyError, IndexError, TypeError, ValueError):
        exists = False

    return exists


def invert(obj):
    """Creates an object composed of the inverted keys and values of the given
    object.

    Args:
        obj (dict): Dict to invert.
        multivalue (bool, optional): Whether to return inverted values as
            lists. Defaults to ``False``.

    Returns:
        dict: Inverted dict.

    Example:

        >>> results = invert({'a': 1, 'b': 2, 'c': 3})
        >>> results == {1: 'a', 2: 'b', 3: 'c'}
        True

    Note:
        Assumes `obj` values are hashable as ``dict`` keys.

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0
        Added ``multivalue`` argument.

    .. versionchanged:: 4.0.0
        Moved ``multivalue=True`` functionality to :func:`invert_by`.
    """
    return dict((value, key) for key, value in iterator(obj))


def invert_by(obj, iteratee=None):
    """This method is like :func:`invert` except that the inverted object is
    generated from the results of running each element of object thru iteratee.
    The corresponding inverted value of each inverted key is a list of keys
    responsible for generating the inverted value. The iteratee is invoked with
    one argument: ``(value)``.

    Args:
        obj (dict): Object to invert.

    Returns:
        dict: Inverted dict.

    Example:

        >>> obj = {'a': 1, 'b': 2, 'c': 1}
        >>> results = invert_by(obj)  # {1: ['a', 'c'], 2: ['b']}
        >>> set(results[1]) == set(['a', 'c'])
        True
        >>> set(results[2]) == set(['b'])
        True
        >>> results2 = invert_by(obj, lambda value: 'group' + str(value))
        >>> results2['group1'] == results[1]
        True
        >>> results2['group2'] == results[2]
        True

    Note:
        Assumes `obj` values are hashable as ``dict`` keys.

    .. versionadded:: 4.0.0
    """
    callback = pyd.iteratee(iteratee)
    result = {}

    for key, value in iterator(obj):
        result.setdefault(callback(value), []).append(key)

    return result


def invoke(obj, path, *args, **kargs):
    """Invokes the method at path of object.

    Args:
        obj (dict): The object to query.
        path (list|str): The path of the method to invoke.
        args (optional): Arguments to pass to method call.
        kargs (optional): Keyword arguments to pass to method call.

    Returns:
        mixed: Result of the invoked method.

    Example:

        >>> obj = {'a': [{'b': {'c': [1, 2, 3, 4]}}]}
        >>> invoke(obj, 'a[0].b.c.pop', 1)
        2
        >>> obj
        {'a': [{'b': {'c': [1, 3, 4]}}]}

    .. versionadded:: 1.0.0
    """
    paths = to_path(path)
    target_path = pyd.initial(paths)
    method_name = pyd.last(paths)

    try:
        method = getattr(get(obj, target_path), method_name)
    except AttributeError:
        ret = None
    else:
        ret = method(*args, **kargs)

    return ret


def keys(obj):
    """Creates a list composed of the keys of `obj`.

    Args:
        obj (mixed): Object to extract keys from.

    Returns:
        list: List of keys.

    Example:

        >>> keys([1, 2, 3])
        [0, 1, 2]
        >>> set(keys({'a': 1, 'b': 2, 'c': 3})) == set(['a', 'b', 'c'])
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Added ``keys_in`` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``keys_in``.
    """
    return [key for key, _ in iterator(obj)]


def map_keys(obj, iteratee=None):
    """The opposite of :func:`map_values`, this method creates an object with
    the same values as object and keys generated by running each own enumerable
    string keyed property of object thru iteratee. The iteratee is invoked with
    three arguments: ``(value, key, object)``.

    Args:
        obj (list|dict): Object to map.
        iteratee (mixed): Iteratee applied per iteration.

    Returns:
        list|dict: Results of running `obj` through `iteratee`.

    Example:

        >>> callback = lambda value, key: key * 2
        >>> results = map_keys({'a': 1, 'b': 2, 'c': 3}, callback)
        >>> results == {'aa': 1, 'bb': 2, 'cc': 3}
        True

    .. versionadded:: 3.3.0
    """
    return dict((result, value)
                for result, value, _, _ in iteriteratee(obj, iteratee))


def map_values(obj, iteratee=None):
    """Creates an object with the same keys as object and values generated by
    running each string keyed property of object thru iteratee. The iteratee is
    invoked with three arguments: ``(value, key, object)``.

    Args:
        obj (list|dict): Object to map.
        iteratee (mixed): Iteratee applied per iteration.

    Returns:
        list|dict: Results of running `obj` through `iteratee`.

    Example:

        >>> results = map_values({'a': 1, 'b': 2, 'c': 3}, lambda x: x * 2)
        >>> results == {'a': 2, 'b': 4, 'c': 6}
        True
        >>> results = map_values({'a': 1, 'b': {'d': 4}, 'c': 3}, {'d': 4})
        >>> results == {'a': False, 'b': True, 'c': False}
        True

    .. versionadded:: 1.0.0
    """
    return dict((key, result)
                for result, _, key, _ in iteriteratee(obj, iteratee))


def map_values_deep(obj, iteratee=None, property_path=NoValue):
    """Map all non-object values in `obj` with return values from `iteratee`.
    The iteratee is invoked with two arguments: ``(obj_value, property_path)``
    where ``property_path`` contains the list of path keys corresponding to the
    path of ``obj_value``.

    Args:
        obj (list|dict): Object to map.
        iteratee (function): Iteratee applied to each value.

    Returns:
        mixed: The modified object.

    Warning:
        `obj` is modified in place.

    Example:

        >>> x = {'a': 1, 'b': {'c': 2}}
        >>> y = map_values_deep(x, lambda val: val * 2)
        >>> y == {'a': 2, 'b': {'c': 4}}
        True
        >>> z = map_values_deep(x, lambda val, props: props)
        >>> z == {'a': ['a'], 'b': {'c': ['b', 'c']}}
        True

    .. versionadded: 2.2.0

    .. versionchanged:: 3.0.0
        Allow iteratees to accept partial arguments.

    .. versionchanged:: 4.0.0
        Renamed from ``deep_map_values`` to ``map_values_deep``.
    """
    properties = to_path(property_path)

    if pyd.is_object(obj):
        deep_iteratee = (
            lambda value, key: map_values_deep(value,
                                               iteratee,
                                               pyd.flatten([properties, key])))
        return assign(obj, map_values(obj, deep_iteratee))
    else:
        return callit(iteratee, obj, properties)


def merge(obj, *sources):
    """Recursively merges properties of the source object(s) into the
    destination object. Subsequent sources will overwrite property assignments
    of previous sources.

    Args:
        obj (dict): Destination object to merge source(s) into.
        sources (dict): Source objects to merge from. subsequent sources
            overwrite previous ones.

    Returns:
        dict: Merged object.

    Warning:
        `obj` is modified in place.

    Example:

        >>> obj = {'a': 2}
        >>> obj2 = merge(obj, {'a': 1}, {'b': 2, 'c': 3}, {'d': 4})
        >>> obj2 == {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        True
        >>> obj is obj2
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.3.2
        Apply :func:`clone_deep` to each `source` before assigning to `obj`.

    .. versionchanged:: 2.3.2
        Allow `iteratee` to be passed by reference if it is the last positional
        argument.

    .. versionchanged:: 4.0.0
        Moved iteratee argument to :func:`merge_with`.
    """
    return merge_with(obj, *sources)


def merge_with(obj, *sources, **kargs):
    """This method is like :func:`merge` except that it accepts customizer
    which is invoked to produce the merged values of the destination and source
    properties. If customizer returns ``None``, merging is handled by this
    method instead. The customizer is invoked with five arguments:
    ``(obj_value, src_value, key, obj, source)``.

    Args:
        obj (dict): Destination object to merge source(s) into.
        sources (dict): Source objects to merge from. subsequent sources
            overwrite previous ones.

    Keyword Args:
        iteratee (function, optional): Iteratee function to handle merging
            (must be passed in as keyword argument).

    Returns:
        dict: Merged object.

    Warning:
        `obj` is modified in place.

    Example:

        >>> cbk = lambda obj_val, src_val: obj_val + src_val
        >>> obj1 = {'a': [1], 'b': [2]}
        >>> obj2 = {'a': [3], 'b': [4]}
        >>> res = merge_with(obj1, obj2, cbk)
        >>> obj1 == {'a': [1, 3], 'b': [2, 4]}
        True

    .. versionadded:: 4.0.0
    """
    sources = list(sources)
    _clone = kargs.get('_clone', True)
    iteratee = kargs.get('iteratee')
    setter = kargs.get('_setter', base_set)

    if iteratee is None and callable(sources[-1]):
        iteratee = sources.pop()

    if callable(iteratee):
        argcount = getargcount(iteratee, maxargs=5)
        cbk = partial(callit, iteratee, argcount=argcount)
    else:
        cbk = None

    for source in sources:
        # Don't re-clone if we've already cloned before.
        if _clone:
            source = copy.deepcopy(source)

        for key, src_value in iterator(source):
            obj_value = base_get(obj, key, default=None)
            all_sequences = all([isinstance(src_value, list),
                                 isinstance(obj_value, list)])
            all_mappings = all([isinstance(src_value, dict),
                                isinstance(obj_value, dict)])

            if cbk:
                _result = cbk(obj_value, src_value, key, obj, source)
            else:
                _result = None

            if _result is not None:
                result = _result
            elif all_sequences or all_mappings:
                result = merge_with(obj_value,
                                    src_value,
                                    _clone=False,
                                    _setter=setter)
            else:
                result = src_value

            setter(obj, key, result)

    return obj


def omit(obj, *properties):
    """The opposite of :func:`pick`. This method creates an object composed of
    the property paths of `obj` that are not omitted.

    Args:
        obj (mixed): Object to process.
        *properties (str): Property values to omit.

    Returns:
        dict: Results of omitting properties.

    Example:

        >>> omit({'a': 1, 'b': 2, 'c': 3}, 'b', 'c') == {'a': 1}
        True
        >>> omit({'a': 1, 'b': 2, 'c': 3 }, ['a', 'c']) == {'b': 2}
        True
        >>> omit([1, 2, 3, 4], 0, 3) == {1: 2, 2: 3}
        True
        >>> omit({'a': {'b': {'c': 'd'}}}, 'a.b.c') == {'a': {'b': {}}}
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee argument to :func:`omit_by`.

    .. versionchanged:: 4.2.0
        Support deep paths.
    """
    return omit_by(obj, pyd.flatten(properties))


def omit_by(obj, iteratee=None):
    """The opposite of :func:`pick_by`. This method creates an object composed
    of the string keyed properties of object that predicate doesn't return
    truthy for. The predicate is invoked with two arguments: ``(value, key)``.

    Args:
        obj (mixed): Object to process.
        iteratee (mixed, optional): Iteratee used to determine which properties
            to omit.

    Returns:
        dict: Results of omitting properties.

    Example:

        >>> omit_by({'a': 1, 'b': '2', 'c': 3}, lambda v: isinstance(v, int))
        {'b': '2'}

    .. versionadded:: 4.0.0

    .. versionchanged:: 4.2.0
        Support deep paths for `iteratee`.
    """
    if not callable(iteratee):
        paths = pyd.map_(iteratee, to_path)

        if any(len(path) > 1 for path in paths):
            cloned = clone_deep(obj)
        else:
            cloned = to_dict(obj)

        def _unset(obj, path):
            pyd.unset(obj, path)
            return obj

        ret = pyd.reduce_(paths, _unset, cloned)
    else:
        argcount = getargcount(iteratee, maxargs=2)

        ret = dict((key, value) for key, value in iterator(obj)
                   if not callit(iteratee, value, key, argcount=argcount))

    return ret


def parse_int(value, radix=None):
    """Converts the given `value` into an integer of the specified `radix`. If
    `radix` is falsey, a radix of ``10`` is used unless the `value` is a
    hexadecimal, in which case a radix of 16 is used.

    Args:
        value (mixed): Value to parse.
        radix (int, optional): Base to convert to.

    Returns:
        mixed: Integer if parsable else ``None``.

    Example:

        >>> parse_int('5')
        5
        >>> parse_int('12', 8)
        10
        >>> parse_int('x') is None
        True

    .. versionadded:: 1.0.0
    """
    if not radix and pyd.is_string(value):
        try:
            # Check if value is hexadcimal and if so use base-16 conversion.
            int(value, 16)
        except ValueError:
            pass
        else:
            radix = 16

    if not radix:
        radix = 10

    try:
        # NOTE: Must convert value to string when supplying radix to int().
        # Dropping radix arg when 10 is needed to allow floats to parse
        # correctly.
        args = (value,) if radix == 10 else (to_string(value), radix)
        parsed = int(*args)
    except (ValueError, TypeError):
        parsed = None

    return parsed


def pick(obj, *properties):
    """Creates an object composed of the picked object properties.

    Args:
        obj (list|dict): Object to pick from.
        properties (str): Property values to pick.

    Returns:
        dict: Dict containg picked properties.

    Example:

        >>> pick({'a': 1, 'b': 2, 'c': 3}, 'a', 'b') == {'a': 1, 'b': 2}
        True

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee argument to :func:`pick_by`.
    """
    return pick_by(obj, pyd.flatten(properties))


def pick_by(obj, iteratee=None):
    """Creates an object composed of the object properties predicate returns
    truthy for. The predicate is invoked with two arguments: ``(value, key)``.

    Args:
        obj (list|dict): Object to pick from.
        iteratee (mixed, optional): Iteratee used to determine which properties
            to pick.

    Returns:
        dict: Dict containg picked properties.

    Example:

        >>> obj = {'a': 1, 'b': '2', 'c': 3 }
        >>> pick_by(obj, lambda v: isinstance(v, int)) == {'a': 1, 'c': 3}
        True

    .. versionadded:: 4.0.0
    """
    obj = to_dict(obj)

    if not callable(iteratee):
        paths = iteratee if iteratee is not None else []

        def iteratee(value, path):  # pylint: disable=function-redefined
            return has(obj, path)

        argcount = 2
    else:
        paths = keys(obj)
        argcount = getargcount(iteratee, maxargs=2)

    result = {}

    for path in paths:
        value = get(obj, path)

        if callit(iteratee, value, path, argcount=argcount):
            set_(result, path, value)

    return result


def rename_keys(obj, key_map):
    """Rename the keys of `obj` using `key_map` and return new object.

    Args:
        obj (dict): Object to rename.
        key_map (dict): Renaming map whose keys correspond to existing keys in
            `obj` and whose values are the new key name.

    Returns:
        dict: Renamed `obj`.

    Example:

        >>> obj = rename_keys({'a': 1, 'b': 2, 'c': 3}, {'a': 'A', 'b': 'B'})
        >>> obj == {'A': 1, 'B': 2, 'c': 3}
        True

    .. versionadded:: 2.0.0
    """
    return dict((key_map.get(key, key), value)
                for key, value in iteritems(obj))


def set_(obj, path, value):
    """Sets the value of an object described by `path`. If any part of the
    object path doesn't exist, it will be created.

    Args:
        obj (list|dict): Object to modify.
        path (str | list): Target path to set value to.
        value (mixed): Value to set.

    Returns:
        mixed: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> set_({}, 'a.b.c', 1)
        {'a': {'b': {'c': 1}}}
        >>> set_({}, 'a.0.c', 1)
        {'a': {'0': {'c': 1}}}
        >>> set_([1, 2], '[2][0]', 1)
        [1, 2, [1]]
        >>> set_({}, 'a.b[0].c', 1)
        {'a': {'b': [{'c': 1}]}}

    .. versionadded:: 2.2.0

    .. versionchanged:: 3.3.0
        Added :func:`set_` as main definition and :func:`deep_set` as alias.

    .. versionchanged:: 4.0.0

        - Modify `obj` in place.
        - Support creating default path values as ``list`` or ``dict`` based on
          whether key or index substrings are used.
        - Remove alias ``deep_set``.
    """
    return set_with(obj, path, value)


def set_with(obj, path, value, customizer=None):
    """This method is like :func:`set_` except that it accepts customizer which
    is invoked to produce the objects of path. If customizer returns undefined
    path creation is handled by the method instead. The customizer is invoked
    with three arguments: ``(nested_value, key, nested_object)``.

    Args:
        obj (list|dict): Object to modify.
        path (str | list): Target path to set value to.
        value (mixed): Value to set.
        customizer (function, optional): The function to customize assigned
            values.

    Returns:
        mixed: Modified `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> set_with({}, '[0][1]', 'a', lambda: {})
        {0: {1: 'a'}}

    .. versionadded:: 4.0.0

    .. versionchanged:: 4.3.1
        Fixed bug where a callable `value` was called when being set.
    """
    return update_with(obj, path, pyd.constant(value), customizer=customizer)


def to_boolean(obj, true_values=('true', '1'), false_values=('false', '0')):
    """Convert `obj` to boolean. This is not like the builtin ``bool``
    function. By default commonly considered strings values are converted to
    their boolean equivalent, i.e., ``'0'`` and ``'false'`` are converted to
    ``False`` while ``'1'`` and ``'true'`` are converted to ``True``. If a
    string value is provided that isn't recognized as having a common boolean
    conversion, then the returned value is ``None``. Non-string values of `obj`
    are converted using ``bool``. Optionally, `true_values` and `false_values`
    can be overridden but each value must be a string.

    Args:
        obj (mixed): Object to convert.
        true_values (tuple, optional): Values to consider ``True``. Each value
            must be a string. Comparision is case-insensitive. Defaults to
            ``('true', '1')``.
        false_values (tuple, optional): Values to consider ``False``. Each
            value must be a string. Comparision is case-insensitive. Defaults
            to ``('false', '0')``.

    Returns:
        bool: Boolean value of `obj`.

    Example:

        >>> to_boolean('true')
        True
        >>> to_boolean('1')
        True
        >>> to_boolean('false')
        False
        >>> to_boolean('0')
        False
        >>> assert to_boolean('a') is None

    .. versionadded:: 3.0.0
    """
    if pyd.is_string(obj):
        obj = obj.strip()

        def boolean_match(text, vals):
            if text.lower() in [val.lower() for val in vals]:
                return True
            else:
                return re.match('|'.join(vals), text)

        if true_values and boolean_match(obj, true_values):
            value = True
        elif false_values and boolean_match(obj, false_values):
            value = False
        else:
            value = None
    else:
        value = bool(obj)

    return value


def to_dict(obj):
    """Convert `obj` to ``dict`` by creating a new ``dict`` using `obj` keys
    and values.

    Args:
        obj: (mixed): Object to convert.

    Returns:
        dict: Object converted to ``dict``.

    Example:

        >>> obj = {'a': 1, 'b': 2}
        >>> obj2 = to_dict(obj)
        >>> obj2 == obj
        True
        >>> obj2 is not obj
        True

    .. versionadded:: 3.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``to_plain_object``.

    .. versionchanged:: 4.2.0
        Use ``pydash.helpers.iterator`` to generate key/value pairs.
    """
    return dict(tuple(iterator(obj)))


def to_integer(obj):
    """Converts `obj` to an integer.

    Args:
        obj (str|int|float): Object to convert.

    Returns:
        int: Converted integer or ``0`` if it can't be converted.

    Example:

        >>> to_integer(3.2)
        3
        >>> to_integer('3.2')
        3
        >>> to_integer('3.9')
        3
        >>> to_integer('invalid')
        0

    .. versionadded:: 4.0.0
    """
    try:
        # Convert to float first to handle converting floats as string since
        # int('1.1') would fail but this wouldn't.
        num = int(float(obj))
    except (ValueError, TypeError):
        num = 0

    return num


def to_list(obj, split_strings=True):
    """ Converts a obj, an iterable or a single item to a list.

    Args:
        obj (mixed): Object to convert item or wrap.
        split_strings (bool, optional): Whether to split strings into single
            chars. Defaults to ``True``.

    Returns:
        list: Converted obj or wrapped item.

    Example:

        >>> results = to_list({'a': 1, 'b': 2, 'c': 3})
        >>> assert set(results) == set([1, 2, 3])

        >>> to_list((1, 2, 3, 4))
        [1, 2, 3, 4]

        >>> to_list(1)
        [1]

        >>> to_list([1])
        [1]

        >>> to_list(a for a in [1, 2, 3])
        [1, 2, 3]

        >>> to_list('cat')
        ['c', 'a', 't']

        >>> to_list('cat', split_strings=False)
        ['cat']

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.3.0

        - Wrap non-iterable items in a list.
        - Convert other iterables to list.
        - Byte objects are returned as single character strings in Python 3.
    """
    if isinstance(obj, list):
        return obj[:]
    elif isinstance(obj, dict):
        return obj.values()
    elif not split_strings and (isinstance(obj, string_types) or
                                isinstance(obj, bytes)):
        return [obj]
    elif split_strings and isinstance(obj, bytes):
        # in python3 iterating over bytes gives integers instead of strings
        return list(chr(c) if isinstance(c, int) else c for c in obj)
    else:
        try:
            return list(obj)

        except TypeError:
            return [obj]


def to_number(obj, precision=0):
    """Convert `obj` to a number. All numbers are retuned as ``float``. If
    precision is negative, round `obj` to the nearest positive integer place.
    If `obj` can't be converted to a number, ``None`` is returned.

    Args:
        obj (str|int|float): Object to convert.
        precision (int, optional): Precision to round number to. Defaults to
            ``0``.

    Returns:
        float: Converted number or ``None`` if can't be converted.

    Example:

        >>> to_number('1234.5678')
        1235.0
        >>> to_number('1234.5678', 4)
        1234.5678
        >>> to_number(1, 2)
        1.0

    .. versionadded:: 3.0.0
    """
    try:
        factor = pow(10, precision)

        if precision < 0:
            # Round down since negative `precision` means we are going to
            # the nearest positive integer place.
            rounder = math.floor
        else:
            rounder = round

        num = rounder(float(obj) * factor) / factor
    except Exception:  # pylint: disable=broad-except
        num = None

    return num


def to_pairs(obj):
    """Creates a two dimensional list of an object's key-value pairs, i.e.
    ``[[key1, value1], [key2, value2]]``.

    Args:
        obj (mixed): Object to process.

    Returns:
        list: Two dimensional list of object's key-value pairs.

    Example:

        >>> to_pairs([1, 2, 3, 4])
        [[0, 1], [1, 2], [2, 3], [3, 4]]
        >>> to_pairs({'a': 1})
        [['a', 1]]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``pairs`` to ``to_pairs``.
    """
    return [[key, value] for key, value in iterator(obj)]


def to_string(obj):
    """Converts an object to string.

    Args:
        obj (mixed): Object to convert.

    Returns:
        str: String representation of `obj`.

    Example:

        >>> to_string(1) == '1'
        True
        >>> to_string(None) == ''
        True
        >>> to_string([1, 2, 3]) == '[1, 2, 3]'
        True
        >>> to_string('a') == 'a'
        True

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Convert ``None`` to empty string.
    """
    if pyd.is_string(obj):
        res = obj
    elif obj is None:
        res = ''
    else:
        res = text_type(obj)
    return res


def transform(obj, iteratee=None, accumulator=None):
    """An alernative to :func:`pydash.collections.reduce`, this method
    transforms `obj` to a new accumulator object which is the result of running
    each of its properties through a iteratee, with each iteratee execution
    potentially mutating the accumulator object. The iteratee is invoked with
    four arguments: ``(accumulator, value, key, object)``. Iteratees may exit
    iteration early by explicitly returning ``False``.

    Args:
        obj (list|dict): Object to process.
        iteratee (mixed): Iteratee applied per iteration.
        accumulator (mixed, optional): Accumulated object. Defaults to
            ``list``.

    Returns:
        mixed: Accumulated object.

    Example:

        >>> transform([1, 2, 3, 4], lambda acc, v, k: acc.append((k, v)))
        [(0, 1), (1, 2), (2, 3), (3, 4)]

    .. versionadded:: 1.0.0
    """
    if iteratee is None:
        iteratee = pyd.identity

    if accumulator is None:
        accumulator = []

    argcount = getargcount(iteratee, maxargs=4)

    walk = (None for key, item in iterator(obj)
            if callit(iteratee,
                      accumulator,
                      item,
                      key,
                      obj,
                      argcount=argcount) is False)
    next(walk, None)

    return accumulator


def update(obj, path, updater):
    """This method is like :func:`set_` except that accepts updater to produce
    the value to set. Use :func:`update_with` to customize path creation. The
    updater is invoked with one argument: ``(value)``.

    Args:
        obj (list|dict): Object to modify.
        path (str|list): A string or list of keys that describe the object path
            to modify.
        updater (function): Function that returns updated value.

    Returns:
        mixed: Updated `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> update({}, ['a', 'b'], lambda value: value)
        {'a': {'b': None}}
        >>> update([], [0, 0], lambda value: 1)
        [[1]]

    .. versionadded:: 4.0.0
    """
    return update_with(obj, path, updater)


def update_with(obj, path, updater, customizer=None):
    """This method is like :func:`update` except that it accepts customizer
    which is invoked to produce the objects of path. If customizer returns
    ``None``, path creation is handled by the method instead. The customizer is
    invoked with three arguments: ``(nested_value, key, nested_object)``.

    Args:
        obj (list|dict): Object to modify.
        path (str|list): A string or list of keys that describe the object path
            to modify.
        updater (function): Function that returns updated value.
        customizer (function, optional): The function to customize assigned
            values.

    Returns:
        mixed: Updated `obj`.

    Warning:
        `obj` is modified in place.

    Example:

        >>> update_with({}, '[0][1]', lambda: 'a', lambda: {})
        {0: {1: 'a'}}

    .. versionadded:: 4.0.0
    """
    if not callable(updater):
        updater = pyd.constant(updater)

    if customizer is not None and not callable(customizer):
        call_customizer = partial(callit, clone, customizer, argcount=1)
    elif customizer:
        call_customizer = partial(callit, customizer,
                                  argcount=getargcount(customizer, maxargs=3))
    else:
        call_customizer = None

    default_type = dict if isinstance(obj, dict) else list
    tokens = to_path_tokens(path)

    if not pyd.is_list(tokens):  # pragma: no cover
        tokens = [tokens]

    last_key = pyd.last(tokens)

    if isinstance(last_key, PathToken):
        last_key = last_key.key

    target = obj

    for idx, token in enumerate(pyd.initial(tokens)):
        if isinstance(token, PathToken):
            key = token.key
            default_factory = pyd.get(tokens,
                                      [idx + 1, 'default_factory'],
                                      default=default_type)
        else:
            key = token
            default_factory = default_type

        obj_val = base_get(target, key, default=None)
        path_obj = None

        if call_customizer:
            path_obj = call_customizer(obj_val, key, target)

        if path_obj is None:
            path_obj = default_factory()

        base_set(target, key, path_obj, allow_override=False)

        try:
            target = target[key]
        except TypeError:  # pragma: no cover
            target = target[int(key)]

    value = base_get(target, last_key, default=None)
    base_set(target, last_key, callit(updater, value))

    return obj


def unset(obj, path):
    """Removes the property at `path` of `obj`.

    Note:
        Only ``list``, ``dict``, or objects with a ``pop()`` method can be
        unset by this function.

    Args:
        obj (mixed): The object to modify.
        path (mixed): The path of the property to unset.

    Returns:
        bool: Whether the property was deleted.

    Warning:
        `obj` is modified in place.

    Example:

        >>> obj = {'a': [{'b': {'c': 7}}]}
        >>> unset(obj, 'a[0].b.c')
        True
        >>> obj
        {'a': [{'b': {}}]}
        >>> unset(obj, 'a[0].b.c')
        False
    """
    tokens = to_path_tokens(path)

    if not pyd.is_list(tokens):  # pragma: no cover
        tokens = [tokens]

    last_key = pyd.last(tokens)

    if isinstance(last_key, PathToken):
        last_key = last_key.key

    target = obj

    for idx, token in enumerate(pyd.initial(tokens)):
        if isinstance(token, PathToken):
            key = token.key
        else:
            key = token

        try:
            try:
                target = target[key]
            except TypeError:
                target = target[int(key)]
        except Exception:
            target = NoValue

        if target is NoValue:
            break

    did_unset = False

    if target is not NoValue:
        try:
            try:
                target.pop(last_key)
                did_unset = True
            except TypeError:
                target.pop(int(last_key))
                did_unset = True
        except Exception:
            pass

    return did_unset


def values(obj):
    """Creates a list composed of the values of `obj`.

    Args:
        obj (mixed): Object to extract values from.

    Returns:
        list: List of values.

    Example:

        >>> results = values({'a': 1, 'b': 2, 'c': 3})
        >>> set(results) == set([1, 2, 3])
        True
        >>> values([2, 4, 6, 8])
        [2, 4, 6, 8]

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Added ``values_in`` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``values_in``.
    """
    return [value for _, value in iterator(obj)]


#
# Utility methods not a part of the main API
#


def base_clone(value, is_deep=False, customizer=None, key=None, obj=None,
               _cloned=False):
    """Base clone function that supports deep clone and customizer callback."""
    clone_by = copy.deepcopy if is_deep else copy.copy
    result = None

    if callable(customizer) and not _cloned:
        argcount = getargcount(customizer, maxargs=4)
        cbk = partial(callit, customizer, argcount=argcount)
    elif _cloned:
        cbk = customizer
    else:
        cbk = None

    if cbk:
        result = cbk(value, key, value)

    if result is not None:
        return result

    if not _cloned:
        result = clone_by(value)
    else:
        result = value

    if cbk:
        for key, subvalue in iterator(value):
            if is_deep:
                val = base_clone(subvalue, is_deep, cbk, key, value,
                                 _cloned=True)
            else:
                val = cbk(subvalue, key, value)

            if val is not None:
                result[key] = val

    return result



# -*- coding: utf-8 -*-

from copy import deepcopy

import pydash as _

from .fixtures import parametrize


pydash_methods = _.filter_(dir(_), lambda m: callable(getattr(_, m, None)))


def test_chaining_methods():
    chain = _.chain([])

    for method in dir(_):
        if not callable(method):
            continue

        chained = getattr(chain, method)
        assert chained.method is getattr(_, method)


@parametrize('value,methods', [
    ([1, 2, 3, 4], [('without', (2, 3)),
                    ('reject', (lambda x: x > 1,))])
])
def test_chaining(value, methods):
    expected = deepcopy(value)
    actual = _.chain(deepcopy(value))

    for method, args in methods:
        expected = getattr(_, method)(expected, *args)
        actual = getattr(actual, method)(*args)

    assert actual.value() == expected


def test_chaining_invalid_method():
    raised = False

    try:
        _.chain([]).foobar
    except _.InvalidMethod:
        raised = True

    assert raised


def test_chaining_lazy():
    tracker = {'called': False}

    def interceptor(value):
        tracker['called'] = True
        return value.pop()

    chain = _.chain([1, 2, 3, 4, 5]).initial().tap(interceptor)

    assert not tracker['called']

    chain = chain.last()

    assert not tracker['called']

    result = chain.value()

    assert tracker['called']
    assert result == 3


def test_chaining_late_value():
    square_sum = _.chain().power(2).sum()
    assert square_sum([1, 2, 3, 4]) == 30


def test_chaining_late_value_reuse():
    square_sum = _.chain().power(2).sum()
    assert square_sum([1, 2, 3, 4]) == 30
    assert square_sum([2]) == 4


def test_chaining_late_value_override():
    square_sum = _.chain([1, 2, 3, 4]).power(2).sum()
    assert square_sum([5, 6, 7, 8]) == 174


def test_chaining_plant():
    value = [1, 2, 3, 4]
    square_sum1 = _.chain(value).power(2).sum()

    def root_value(wrapper):
        if isinstance(wrapper._value, _.chaining.ChainWrapper):
            return root_value(wrapper._value)
        return wrapper._value

    assert root_value(square_sum1._value) == value

    test_value = [5, 6, 7, 8]
    square_sum2 = square_sum1.plant(test_value)

    assert root_value(square_sum1._value) == value
    assert root_value(square_sum2._value) == test_value

    assert square_sum1.value() == 30
    assert square_sum2.value() == 174


def test_chaining_commit():
    chain = _.chain([1, 2, 3, 4]).power(2).sum()
    committed = chain.commit()

    assert chain is not committed
    assert chain.value() == committed.value()


def test_dash_instance_chaining():
    value = [1, 2, 3, 4]
    from__ = _._(value).without(2, 3).reject(lambda x: x > 1)
    from_chain = _.chain(value).without(2, 3).reject(lambda x: x > 1)

    assert from__.value() == from_chain.value()


def test_dash_instance_methods():
    assert pydash_methods

    for method in pydash_methods:
        assert getattr(_._, method) is getattr(_, method)


def test_dash_suffixed_method_aliases():
    methods = _.filter_(pydash_methods, lambda m: m.endswith('_'))
    assert methods

    for method in methods:
        assert getattr(_._, method[:-1]) is getattr(_, method)


def test_dash_method_call():
    value = [1, 2, 3, 4, 5]
    assert _._.initial(value) == _.initial(value)


def test_dash_alias():
    assert _.py_ is _._


@parametrize('case,expected', [
    ([1, 2, 3], '[1, 2, 3]'),
])
def test_chaining_value_to_string(case, expected):
    assert _.chain(case).to_string() == expected


@parametrize('value,interceptor,expected', [
    ([1, 2, 3, 4, 5], lambda value: value.pop(), 3)
])
def test_tap(value, interceptor, expected):
    actual = _.chain(value).initial().tap(interceptor).last().value()
    assert actual == expected


@parametrize('value,func,expected', [
    ([1, 2, 3, 4, 5], lambda value: [sum(value)], 10)
])
def test_thru(value, func, expected):
    assert _.chain(value).initial().thru(func).last().value()



# -*- coding: utf-8 -*-
"""Predicate functions that return boolean evaluations of objects.

.. versionadded:: 2.0.0
"""

from __future__ import absolute_import

import datetime
from itertools import islice
import json
import operator
import re
from types import BuiltinFunctionType

import pydash as pyd
from .helpers import NoValue, callit, iterator
from ._compat import (
    builtins,
    integer_types,
    izip,
    number_types,
    string_types
)


__all__ = (
    'eq',
    'gt',
    'gte',
    'lt',
    'lte',
    'in_range',
    'is_associative',
    'is_blank',
    'is_boolean',
    'is_builtin',
    'is_date',
    'is_decreasing',
    'is_dict',
    'is_empty',
    'is_equal',
    'is_equal_with',
    'is_error',
    'is_even',
    'is_float',
    'is_function',
    'is_increasing',
    'is_indexed',
    'is_instance_of',
    'is_integer',
    'is_iterable',
    'is_json',
    'is_list',
    'is_match',
    'is_match_with',
    'is_monotone',
    'is_nan',
    'is_negative',
    'is_none',
    'is_number',
    'is_object',
    'is_odd',
    'is_positive',
    'is_reg_exp',
    'is_set',
    'is_strictly_decreasing',
    'is_strictly_increasing',
    'is_string',
    'is_tuple',
    'is_zero',
)


RegExp = type(re.compile(''))


def eq(value, other):
    """Checks if :attr:`value` is equal to :attr:`other`.

    Args:
        value (mixed): Value to compare.
        other (mixed): Other value to compare.

    Returns:
        bool: Whether :attr:`value` is equal to :attr:`other`.

    Example:

        >>> eq(None, None)
        True
        >>> eq(None, '')
        False
        >>> eq('a', 'a')
        True
        >>> eq(1, str(1))
        False

    .. versionadded:: 4.0.0
    """
    return value is other


def gt(value, other):
    """Checks if `value` is greater than `other`.

    Args:
        value (number): Value to compare.
        other (number): Other value to compare.

    Returns:
        bool: Whether `value` is greater than `other`.

    Example:

        >>> gt(5, 3)
        True
        >>> gt(3, 5)
        False
        >>> gt(5, 5)
        False

    .. versionadded:: 3.3.0
    """
    return value > other


def gte(value, other):
    """Checks if `value` is greater than or equal to `other`.

    Args:
        value (number): Value to compare.
        other (number): Other value to compare.

    Returns:
        bool: Whether `value` is greater than or equal to `other`.

    Example:

        >>> gte(5, 3)
        True
        >>> gte(3, 5)
        False
        >>> gte(5, 5)
        True

    .. versionadded:: 3.3.0
    """
    return value >= other


def lt(value, other):
    """Checks if `value` is less than `other`.

    Args:
        value (number): Value to compare.
        other (number): Other value to compare.

    Returns:
        bool: Whether `value` is less than `other`.

    Example:

        >>> lt(5, 3)
        False
        >>> lt(3, 5)
        True
        >>> lt(5, 5)
        False

    .. versionadded:: 3.3.0
    """
    return value < other


def lte(value, other):
    """Checks if `value` is less than or equal to `other`.

    Args:
        value (number): Value to compare.
        other (number): Other value to compare.

    Returns:
        bool: Whether `value` is less than or equal to `other`.

    Example:

        >>> lte(5, 3)
        False
        >>> lte(3, 5)
        True
        >>> lte(5, 5)
        True

    .. versionadded:: 3.3.0
    """
    return value <= other


def in_range(value, start=0, end=None):
    """Checks if `value` is between `start` and up to but not including `end`.
    If `end` is not specified it defaults to `start` with `start` becoming
    ``0``.

    Args:

        value (int|float): Number to check.
        start (int|float, optional): Start of range inclusive. Defaults to
            ``0``.
        end (int|float, optional): End of range exclusive. Defaults to `start`.

    Returns:
        bool: Whether `value` is in range.

    Example:

        >>> in_range(2, 4)
        True
        >>> in_range(4, 2)
        False
        >>> in_range(2, 1, 3)
        True
        >>> in_range(3, 1, 2)
        False
        >>> in_range(2.5, 3.5)
        True
        >>> in_range(3.5, 2.5)
        False

    .. versionadded:: 3.1.0
    """
    if not is_number(value):
        return False

    if not is_number(start):
        start = 0

    if end is None:
        end = start
        start = 0
    elif not is_number(end):
        end = 0

    return start <= value < end


def is_associative(value):
    """Checks if `value` is an associative object meaning that it can be
    accessed via an index or key

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is associative.

    Example:

        >>> is_associative([])
        True
        >>> is_associative({})
        True
        >>> is_associative(1)
        False
        >>> is_associative(True)
        False

    .. versionadded:: 2.0.0
    """
    return hasattr(value, '__getitem__')


def is_blank(text):
    r"""Checks if `text` contains only whitespace characters.

    Args:
        text (str): String to test.

    Returns:
        bool: Whether `text` is blank.

    Example:

        >>> is_blank('')
        True
        >>> is_blank(' \r\n ')
        True
        >>> is_blank(False)
        False

    .. versionadded:: 3.0.0
    """
    try:
        ret = bool(re.match(r'^(\s+)?$', text))
    except TypeError:
        ret = False

    return ret


def is_boolean(value):
    """Checks if `value` is a boolean value.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a boolean.

    Example:

        >>> is_boolean(True)
        True
        >>> is_boolean(False)
        True
        >>> is_boolean(0)
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Added ``is_bool`` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``is_bool``.
    """
    return isinstance(value, bool)


def is_builtin(value):
    """Checks if `value` is a Python builtin function or method.

    Args:
        value (function): Value to check.

    Returns:
        bool: Whether `value` is a Python builtin function or method.

    Example:

        >>> is_builtin(1)
        True
        >>> is_builtin(list)
        True
        >>> is_builtin('foo')
        False

    .. versionadded:: 3.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``is_native``.
    """
    try:
        return isinstance(value, BuiltinFunctionType) or value in builtins
    except TypeError:  # pragma: no cover
        return False


def is_date(value):
    """Check if `value` is a date object.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a date object.

    Example:

        >>> import datetime
        >>> is_date(datetime.date.today())
        True
        >>> is_date(datetime.datetime.today())
        True
        >>> is_date('2014-01-01')
        False

    Note:
        This will also return ``True`` for datetime objects.

    .. versionadded:: 1.0.0
    """
    return isinstance(value, datetime.date)


def is_decreasing(value):
    """Check if `value` is monotonically decreasing.

    Args:
        value (list): Value to check.

    Returns:
        bool: Whether `value` is monotonically decreasing.

    Example:

        >>> is_decreasing([5, 4, 4, 3])
        True
        >>> is_decreasing([5, 5, 5])
        True
        >>> is_decreasing([5, 4, 5])
        False

    .. versionadded:: 2.0.0
    """
    return is_monotone(value, operator.ge)


def is_dict(value):
    """Checks if `value` is a ``dict``.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a ``dict``.

    Example:

        >>> is_dict({})
        True
        >>> is_dict([])
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Added :func:`is_dict` as main definition and made `is_plain_object`` an
        alias.

    .. versionchanged:: 4.0.0
        Removed alias ``is_plain_object``.
    """
    return isinstance(value, dict)


def is_empty(value):
    """Checks if `value` is empty.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is empty.

    Example:

        >>> is_empty(0)
        True
        >>> is_empty(1)
        True
        >>> is_empty(True)
        True
        >>> is_empty('foo')
        False
        >>> is_empty(None)
        True
        >>> is_empty({})
        True

    Note:
        Returns ``True`` for booleans and numbers.

    .. versionadded:: 1.0.0
    """
    return any([is_boolean(value), is_number(value), not value])


def is_equal(value, other):
    """Performs a comparison between two values to determine if they are
    equivalent to each other.

    Args:
        value (list|dict): Object to compare.
        other (list|dict): Object to compare.

    Returns:
        bool: Whether `value` and `other` are equal.

    Example:

        >>> is_equal([1, 2, 3], [1, 2, 3])
        True
        >>> is_equal('a', 'A')
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed :attr:`iteratee` from :func:`is_equal` and added it in
        :func:`is_equal_with`.
    """
    return is_equal_with(value, other, customizer=None)


def is_equal_with(value, other, customizer):
    """This method is like _.isEqual except that it accepts customizer which
    is invoked to compare values. A customizer is provided which will be
    executed to compare values. If the customizer returns ``None``, comparisons
    will be handled by the method instead. The customizer is invoked with two
    arguments: ``(value, other)``.

    Args:
        value (list|dict): Object to compare.
        other (list|dict): Object to compare.
        customizer (mixed, optional): Customizer used to compare values from
            `value` and `other`.

    Returns:
        bool: Whether `value` and `other` are equal.

    Example:

        >>> is_equal_with([1, 2, 3], [1, 2, 3], None)
        True
        >>> is_equal_with('a', 'A', None)
        False
        >>> is_equal_with('a', 'A', lambda a, b: a.lower() == b.lower())
        True

    .. versionadded:: 4.0.0
    """
    # If customizer provided, use it for comparision.
    equal = customizer(value, other) if callable(customizer) else None

    # Return customizer results if anything but None.
    if equal is not None:
        pass
    elif (callable(customizer) and
          type(value) is type(other) and
          isinstance(value, (list, dict)) and
          isinstance(other, (list, dict)) and
          len(value) == len(other)):
        # Walk a/b to determine equality using customizer.
        for key, value in iterator(value):
            if pyd.has(other, key):
                equal = is_equal_with(value, other[key], customizer)
            else:
                equal = False

            if not equal:
                break
    else:
        # Use basic == comparision.
        equal = value == other

    return equal


def is_error(value):
    """Checks if `value` is an ``Exception``.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is an exception.

    Example:

        >>> is_error(Exception())
        True
        >>> is_error(Exception)
        False
        >>> is_error(None)
        False

    .. versionadded:: 1.1.0
    """
    return isinstance(value, Exception)


def is_even(value):
    """Checks if `value` is even.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is even.

    Example:

        >>> is_even(2)
        True
        >>> is_even(3)
        False
        >>> is_even(False)
        False

    .. versionadded:: 2.0.0
    """
    return is_number(value) and value % 2 == 0


def is_float(value):
    """Checks if `value` is a float.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a float.

    Example:

        >>> is_float(1.0)
        True
        >>> is_float(1)
        False

    .. versionadded:: 2.0.0
    """
    return isinstance(value, float)


def is_function(value):
    """Checks if `value` is a function.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is callable.

    Example:

        >>> is_function(list)
        True
        >>> is_function(lambda: True)
        True
        >>> is_function(1)
        False

    .. versionadded:: 1.0.0
    """
    return callable(value)


def is_increasing(value):
    """Check if `value` is monotonically increasing.

    Args:
        value (list): Value to check.

    Returns:
        bool: Whether `value` is monotonically increasing.

    Example:

        >>> is_increasing([1, 3, 5])
        True
        >>> is_increasing([1, 1, 2, 3, 3])
        True
        >>> is_increasing([5, 5, 5])
        True
        >>> is_increasing([1, 2, 4, 3])
        False

    .. versionadded:: 2.0.0
    """
    return is_monotone(value, operator.le)


def is_indexed(value):
    """Checks if `value` is integer indexed, i.e., ``list``, ``str`` or
    ``tuple``.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is integer indexed.

    Example:

        >>> is_indexed('')
        True
        >>> is_indexed([])
        True
        >>> is_indexed(())
        True
        >>> is_indexed({})
        False

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Return ``True`` for tuples.
    """
    return isinstance(value, (list, tuple, string_types))


def is_instance_of(value, types):
    """Checks if `value` is an instance of `types`.

    Args:
        value (mixed): Value to check.
        types (mixed): Types to check against. Pass as ``tuple`` to check if
            `value` is one of multiple types.

    Returns:
        bool: Whether `value` is an instance of `types`.

    Example:

        >>> is_instance_of({}, dict)
        True
        >>> is_instance_of({}, list)
        False

    .. versionadded:: 2.0.0
    """
    return isinstance(value, types)


def is_integer(value):
    """Checks if `value` is a integer.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is an integer.

    Example:

        >>> is_integer(1)
        True
        >>> is_integer(1.0)
        False
        >>> is_integer(True)
        False

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        Added ``is_int`` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``is_int``.
    """
    return is_number(value) and isinstance(value, integer_types)


def is_iterable(value):
    """Checks if `value` is an iterable.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is an iterable.

    Example:

        >>> is_iterable([])
        True
        >>> is_iterable({})
        True
        >>> is_iterable(())
        True
        >>> is_iterable(5)
        False
        >>> is_iterable(True)
        False

    .. versionadded:: 3.3.0
    """
    try:
        iter(value)
    except TypeError:
        return False
    else:
        return True


def is_json(value):
    """Checks if `value` is a valid JSON string.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is JSON.

    Example:

        >>> is_json({})
        False
        >>> is_json('{}')
        True
        >>> is_json({"hello": 1, "world": 2})
        False
        >>> is_json('{"hello": 1, "world": 2}')
        True

    .. versionadded:: 2.0.0
    """
    try:
        json.loads(value)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def is_list(value):
    """Checks if `value` is a list.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a list.

    Example:

        >>> is_list([])
        True
        >>> is_list({})
        False
        >>> is_list(())
        False

    .. versionadded:: 1.0.0
    """
    return isinstance(value, list)


def is_match(obj, source):
    """Performs a partial deep comparison between `obj` and `source` to
    determine if `obj` contains equivalent property values.

    Args:
        obj (list|dict): Object to compare.
        source (list|dict): Object of property values to match.

    Returns:
        bool: Whether `obj` is a match or not.

    Example:

        >>> is_match({'a': 1, 'b': 2}, {'b': 2})
        True
        >>> is_match({'a': 1, 'b': 2}, {'b': 3})
        False
        >>> is_match({'a': [{'b': [{'c': 3, 'd': 4}]}]},\
                     {'a': [{'b': [{'d': 4}]}]})
        True

    .. versionadded:: 3.0.0

    .. versionchanged:: 3.2.0
        Don't compare `obj` and `source` using ``type``. Use ``isinstance``
        exclusively.

    .. versionchanged:: 4.0.0
        Move `iteratee` argument to :func:`is_match_with`.
    """
    return is_match_with(obj, source)


def is_match_with(obj, source, customizer=None,
                  _key=NoValue, _obj=NoValue, _source=NoValue):
    """This method is like :func:`is_match` except that it accepts customizer
    which is invoked to compare values. If customizer returns ``None``,
    comparisons are handled by the method instead. The customizer is invoked
    with five arguments: ``(obj_value, src_value, index|key, obj, source)``.

    Args:
        obj (list|dict): Object to compare.
        source (list|dict): Object of property values to match.
        customizer (mixed, optional): Customizer used to compare values from
            `obj` and `source`.

    Returns:
        bool: Whether `obj` is a match or not.

    Example:

        >>> is_greeting = lambda val: val in ('hello', 'hi')
        >>> customizer = lambda ov, sv: is_greeting(ov) and is_greeting(sv)
        >>> obj = {'greeting': 'hello'}
        >>> src = {'greeting': 'hi'}
        >>> is_match_with(obj, src, customizer)
        True

    .. versionadded:: 4.0.0
    """
    if _obj is NoValue:
        _obj = obj

    if _source is NoValue:
        _source = source

    if not callable(customizer):
        def cbk(obj_value, src_value):
            return obj_value == src_value

        cbk._argcount = 2
    else:
        cbk = customizer

    if (isinstance(obj, dict) and isinstance(source, dict) or
            isinstance(obj, list) and isinstance(source, list) or
            isinstance(obj, tuple) and isinstance(source, tuple)):
        # Set equal to True if source is empty, otherwise, False and then allow
        # deep comparison to determine equality.
        equal = not source

        # Walk a/b to determine equality.
        for key, value in iterator(source):
            try:
                equal = is_match_with(obj[key], value, cbk,
                                      _key=key, _obj=_obj, _source=_source)
            except Exception:  # pylint: disable=broad-except
                equal = False

            if not equal:
                break
    else:
        equal = callit(cbk, obj, source, _key, _obj, _source)

    return equal


def is_monotone(value, op):
    """Checks if `value` is monotonic when `operator` used for comparison.

    Args:
        value (list): Value to check.
        op (function): Operation to used for comparison.

    Returns:
        bool: Whether `value` is monotone.

    Example:

        >>> is_monotone([1, 1, 2, 3], operator.le)
        True
        >>> is_monotone([1, 1, 2, 3], operator.lt)
        False

    .. versionadded:: 2.0.0
    """
    if not is_list(value):
        value = [value]

    search = (False for x, y in izip(value, islice(value, 1, None))
              if not op(x, y))

    return next(search, True)


def is_nan(value):
    """Checks if `value` is not a number.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is not a number.

    Example:

        >>> is_nan('a')
        True
        >>> is_nan(1)
        False
        >>> is_nan(1.0)
        False

    .. versionadded:: 1.0.0
    """
    return not is_number(value)


def is_negative(value):
    """Checks if `value` is negative.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is negative.

    Example:

        >>> is_negative(-1)
        True
        >>> is_negative(0)
        False
        >>> is_negative(1)
        False

    .. versionadded:: 2.0.0
    """
    return is_number(value) and value < 0


def is_none(value):
    """Checks if `value` is `None`.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is ``None``.

    Example:

        >>> is_none(None)
        True
        >>> is_none(False)
        False

    .. versionadded:: 1.0.0
    """
    return value is None


def is_number(value):
    """Checks if `value` is a number.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a number.

    Note:
        Returns ``True`` for ``int``, ``long`` (PY2), ``float``, and
        ``decimal.Decimal``.

    Example:

        >>> is_number(1)
        True
        >>> is_number(1.0)
        True
        >>> is_number('a')
        False

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Added ``is_num`` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``is_num``.
    """
    return not is_boolean(value) and isinstance(value, number_types)


def is_object(value):
    """Checks if `value` is a ``list`` or ``dict``.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is ``list`` or ``dict``.

    Example:

        >>> is_object([])
        True
        >>> is_object({})
        True
        >>> is_object(())
        False
        >>> is_object(1)
        False

    .. versionadded:: 1.0.0
    """
    return isinstance(value, (list, dict))


def is_odd(value):
    """Checks if `value` is odd.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is odd.

    Example:

        >>> is_odd(3)
        True
        >>> is_odd(2)
        False
        >>> is_odd('a')
        False

    .. versionadded:: 2.0.0
    """
    return is_number(value) and value % 2 != 0


def is_positive(value):
    """Checks if `value` is positive.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is positive.

    Example:

        >>> is_positive(1)
        True
        >>> is_positive(0)
        False
        >>> is_positive(-1)
        False

    .. versionadded:: 2.0.0
    """
    return is_number(value) and value > 0


def is_reg_exp(value):
    """Checks if `value` is a ``RegExp`` object.

    Args:
        value (mxied): Value to check.

    Returns:
        bool: Whether `value` is a RegExp object.

    Example:

        >>> is_reg_exp(re.compile(''))
        True
        >>> is_reg_exp('')
        False

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``is_re``.
    """
    return isinstance(value, RegExp)


def is_set(value):
    """Checks if the given value is a set object or not.

    Args:
        value (mixed): Value passed in by the user.

    Returns:
        bool: True if the given value is a set else False.

    Example:

        >>> is_set(set([1, 2]))
        True
        >>> is_set([1, 2, 3])
        False

    .. versionadded:: 4.0.0
    """
    return isinstance(value, set)


def is_strictly_decreasing(value):
    """Check if `value` is strictly decreasing.

    Args:
        value (list): Value to check.

    Returns:
        bool: Whether `value` is strictly decreasing.

    Example:

        >>> is_strictly_decreasing([4, 3, 2, 1])
        True
        >>> is_strictly_decreasing([4, 4, 2, 1])
        False

    .. versionadded:: 2.0.0
    """
    return is_monotone(value, operator.gt)


def is_strictly_increasing(value):
    """Check if `value` is strictly increasing.

    Args:
        value (list): Value to check.

    Returns:
        bool: Whether `value` is strictly increasing.

    Example:

        >>> is_strictly_increasing([1, 2, 3, 4])
        True
        >>> is_strictly_increasing([1, 1, 3, 4])
        False

    .. versionadded:: 2.0.0
    """
    return is_monotone(value, operator.lt)


def is_string(value):
    """Checks if `value` is a string.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a string.

    Example:

        >>> is_string('')
        True
        >>> is_string(1)
        False

    .. versionadded:: 1.0.0
    """
    return isinstance(value, string_types)


def is_tuple(value):
    """Checks if `value` is a tuple.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is a tuple.

    Example:

        >>> is_tuple(())
        True
        >>> is_tuple({})
        False
        >>> is_tuple([])
        False

    .. versionadded:: 3.0.0
    """
    return isinstance(value, tuple)


def is_zero(value):
    """Checks if `value` is ``0``.

    Args:
        value (mixed): Value to check.

    Returns:
        bool: Whether `value` is ``0``.

    Example:

        >>> is_zero(0)
        True
        >>> is_zero(1)
        False

    .. versionadded:: 2.0.0
    """
    return value is 0



# -*- coding: utf-8 -*-

import pydash as _
from .fixtures import parametrize

import pytest


@parametrize('case,expected', [
    ((5, 3), 8),
])
def test_add(case, expected):
    assert _.add(*case) == expected


@parametrize('case,expected', [
    ((4.006,), 5),
    ((6.004, 2), 6.01),
    ((6040, -2), 6100),
    (([4.006, 6.004], 2), [4.01, 6.01]),
])
def test_ceil(case, expected):
    assert _.ceil(*case) == expected


@parametrize('case,expected', [
    ((0, -1, 1), 0),
    ((1, -1, 1), 1),
    ((-1, -1, 1), -1),
    ((1, 1), 1),
    ((5, -1, 1), 1),
    ((-5, -1, 1), -1),
])
def test_clamp(case, expected):
    assert _.clamp(*case) == expected


@parametrize('dividend,divisor,expected', [
    (10, 5, 2.0),
    (None, 1, 1.0),
    (None, None, 1.0),
    (1.5, 3, 0.5),
    (-10, 2, -5.0)
])
def test_divide(dividend, divisor, expected):
    assert _.divide(dividend, divisor) == expected


@parametrize('case,expected', [
    ((4.006,), 4),
    ((0.046, 2), 0.04),
    ((4060, -2), 4000),
    (([4.006, 0.046], 2), [4.0, 0.04]),
])
def test_floor(case, expected):
    assert _.floor(*case) == expected


@parametrize('case,expected', [
    (([1, 2, 3],), 3),
    (({'a': 3, 'b': 2, 'c': 1},), 3),
])
def test_max_(case, expected):
    assert _.max_(*case) == expected


@parametrize('case,expected', [
    (([1, 2, 3],), 3),
    (({'a': 3, 'b': 2, 'c': 1},), 3),
    ((['anaconda', 'bison', 'camel'], lambda x: len(x)), 'anaconda'),
    (([{'name': 'barney', 'age': 36}, {'name': 'fred', 'age': 40}], 'age',),
     {'name': 'fred', 'age': 40}),
    (([{'name': 'barney', 'age': 36}, {'name': 'fred', 'age': 40}],
      lambda chr: chr['age']),
     {'name': 'fred', 'age': 40}),
])
def test_max_by(case, expected):
    assert _.max_by(*case) == expected


@parametrize('collection,default,expected', [
    ([], -1, -1),
    ([1, 2, 3], -1, 3),
    ({}, -1, -1),
    ([], None, None),
    ({}, None, None)
])
def test_max_default(collection, default, expected):
    assert _.max_(collection, default=default) == expected


@parametrize('case,expected', [
    ([1, 2, 3, 4, 5], 3),
    ([0, 0.5, 1], 0.5),
])
def test_mean(case, expected):
    assert _.mean(case) == expected


@parametrize('case,expected', [
    (([1, 2, 3, 4, 5],), 3),
    (([{'b': 4}, {'b': 5}, {'b': 6}], 'b'), 5),
    (([0, 0.5, 1],), 0.5),
    (({'one': {'a': 1}, 'two': {'a': 2}, 'three': {'a': 3}}, 'a'), 2),
])
def test_mean_by(case, expected):
    assert _.mean_by(*case) == expected


@parametrize('case,expected', [
    (([0, 0, 0, 0, 5],), 0),
    (([0, 0, 1, 2, 5],), 1),
    (([0, 0, 1, 2],), 0.5),
    (([0, 0, 1, 2, 3, 4],), 1.5),
])
def test_median(case, expected):
    assert _.median(*case) == expected


@parametrize('case,expected', [
    (([1, 2, 3],), 1),
    (({'a': 3, 'b': 2, 'c': 1},), 1),
])
def test_min_(case, expected):
    assert _.min_(*case) == expected


@parametrize('case,expected', [
    (([1, 2, 3],), 1),
    (({'a': 3, 'b': 2, 'c': 1},), 1),
    ((['anaconda', 'bison', 'cat'], lambda x: len(x)), 'cat'),
    (([{'name': 'barney', 'age': 36}, {'name': 'fred', 'age': 40}], 'age',),
     {'name': 'barney', 'age': 36}),
    (([{'name': 'barney', 'age': 36}, {'name': 'fred', 'age': 40}],
      lambda chr: chr['age']),
     {'name': 'barney', 'age': 36}),
])
def test_min_by(case, expected):
    assert _.min_by(*case) == expected


@parametrize('collection,default,expected', [
    ([], -1, -1),
    ([1, 2, 3], -1, 1),
    ({}, -1, -1),
    ([], None, None),
    ({}, None, None)
])
def test_min_default(collection, default, expected):
    assert _.min_(collection, default=default) == expected


@parametrize('case,expected', [
    (([1, 2, 3, 4, 5], 3), [2, 3, 4]),
    (([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3), [2, 3, 4, 5, 6, 7, 8, 9]),
    (([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 4),
     [2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]),
])
def test_moving_mean(case, expected):
    assert _.moving_mean(*case) == expected


@parametrize('multiplier,multiplicand,expected', [
    (10, 5, 50),
    (None, 1, 1),
    (None, None, 1),
    (1.5, 3, 4.5),
    (-10, 2, -20)
])
def test_multiply(multiplier, multiplicand, expected):
    assert _.multiply(multiplier, multiplicand) == expected


@parametrize('case,expected', [
    ((2, 3), 8),
    ((3, 4), 81),
    (([1, 2, 3, 4, 5], 2), [1, 4, 9, 16, 25]),
    (('junk', 2), None),
])
def test_power(case, expected):
    assert _.power(*case) == expected


@parametrize('case,expected', [
    ((2.51,), 3),
    ((2.499,), 2),
    ((2.499, 2), 2.50),
    (([2.5, 2.499, 2.555], 2), [2.50, 2.50, 2.56]),
    (('junk',), None)
])
def test_round_(case, expected):
    assert _.round_(*case) == expected


@parametrize('case,expected', [
    (([2, 5, 10], 1), [0.2, 0.5, 1]),
    (([1, 2, 5], 1), [0.2, 0.4, 1]),
    (([1, 2, 5], 5), [1, 2, 5]),
    (([1, 2, 5],), [0.2, 0.4, 1]),
])
def test_scale(case, expected):
    assert _.scale(*case) == expected


@parametrize('case,expected', [
    (([0, 0], [5, 5]), 1),
    (([0, 0], [1, 10]), 10),
    (([0, 0], [0, 10]), float('inf')),
    (([0, 0], [10, 0]), 0),
])
def test_slope(case, expected):
    assert _.slope(*case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], (2.0 / 3.0) ** 0.5),
])
def test_std_deviation(case, expected):
    assert _.std_deviation(case) == expected


@parametrize('minuend,subtrahend,expected', [
    (10, 4, 6),
    (-6, -4, -2),
    (4, -10, 14),
    (-10, 4, -14),
    ('10', '5', 5),
    (2, 0.5, 1.5),
    (None, None, 0)
])
def test_subtract(minuend, subtrahend, expected):
    assert _.subtract(minuend, subtrahend) == expected


@parametrize('minuend,subtrahend', [
    ('abs', 4),
    (4, 'abc'),
    ('abs', 'abc'),
    ([1, 2, 3, 4], 4),
])
def test_subtract_exception(minuend, subtrahend):
    with pytest.raises(TypeError):
        _.subtract(minuend, subtrahend)


@parametrize('case,expected', [
    ([1, 2, 3, 4, 5], 15),
    ([0, 14, 0.2], 14.2),
])
def test_sum_(case, expected):
    assert _.sum_(case) == expected


@parametrize('case,expected', [
    (([1, 2, 3, 4, 5], lambda a: a * 2), 30),
    (([{'b': 4}, {'b': 5}, {'b': 6}], 'b'), 15),
    (({'one': {'a': 1}, 'two': {'a': 2}, 'three': {'a': 3}}, 'a'), 6),
])
def test_sum_by(case, expected):
    assert _.sum_by(*case) == expected


@parametrize('case,expected', [
    ([[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[1, 4, 7], [2, 5, 8], [3, 6, 9]]),
])
def test_transpose(case, expected):
    assert _.transpose(case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], 2.0 / 3.0),
])
def test_variance(case, expected):
    assert _.variance(case) == expected


@parametrize('case,expected', [
    (([1, 2, 3],), [-1.225, 0.0, 1.225]),
    (([{'a': 1}, {'a': 2}, {'a': 3}], 'a'), [-1.225, 0.0, 1.225]),
])
def test_zscore(case, expected):
    assert _.map_(_.zscore(*case), lambda v: round(v, 3)) == expected



# -*- coding: utf-8 -*-

import re

import pydash as _
from pydash._compat import urlsplit, parse_qsl
from .fixtures import parametrize


@parametrize('case,expected', [
    ('foo bar baz', 'fooBarBaz'),
    ('foo  bar baz', 'fooBarBaz'),
    ('foo__bar_baz', 'fooBarBaz'),
    ('foo-_bar-_-baz', 'fooBarBaz'),
    ('foo!bar,baz', 'fooBarBaz'),
    ('--foo.bar;baz', 'fooBarBaz'),
    (8, '8'),
    ('', ''),
    (None, ''),
])
def test_camel_case(case, expected):
    assert _.camel_case(case) == expected


@parametrize('case,expected', [
    (('foo',), 'Foo'),
    (('foo bar',), 'Foo bar'),
    (('fOO bar',), 'Foo bar'),
    (('foo Bar',), 'Foo bar'),
    (('foo', False), 'Foo'),
    (('foo bar', False), 'Foo bar'),
    (('fOO bar', False), 'FOO bar'),
    (('foo Bar', False), 'Foo Bar'),
    ((8,), '8'),
    ((' ',), ' '),
    (('',), ''),
    ((None,), ''),
])
def test_capitalize(case, expected):
    assert _.capitalize(*case) == expected


@parametrize('case,expected', [
    (('foobarbaz', 3), ['foo', 'bar', 'baz']),
    (('foobarbazaa', 3), ['foo', 'bar', 'baz', 'aa']),
    (('foo', 4), ['foo']),
    (('foo', 0), ['foo']),
    (('', 3), []),
    (('foo', -2), ['foo']),
    ((None, 0), []),
    ((None, 1), []),
    ((None, -1), []),
])
def test_chop(case, expected):
    assert _.chop(*case) == expected


@parametrize('case,expected', [
    (('foobarbaz', 3), ['foo', 'bar', 'baz']),
    (('foobarbazaa', 3), ['fo', 'oba', 'rba', 'zaa']),
    (('foo', 4), ['foo']),
    (('foo', 0), ['foo']),
    (('', 3), []),
    (('foo', -2), ['foo']),
    ((None, 0), []),
    ((None, 1), []),
    ((None, -1), []),
])
def test_chop_right(case, expected):
    assert _.chop_right(*case) == expected


@parametrize('case,expected', [
    ('  foo bar', 'foo bar'),
    ('  foo  bar', 'foo bar'),
    ('  foo  bar  ', 'foo bar'),
    ('', ''),
    (None, ''),
])
def test_clean(case, expected):
    assert _.clean(case) == expected


@parametrize('case,expected', [
    ('foobar', ['f', 'o', 'o', 'b', 'a', 'r']),
    ('', []),
    (5, ['5']),
    (-5.6, ['-', '5', '.', '6']),
    (None, []),
])
def test_chars(case, expected):
    assert _.chars(case) == expected


@parametrize('case,expected', [
    (('foobar', 'o'), 2),
    (('foobar', 'oo'), 1),
    (('foobar', 'ooo'), 0),
    (('', 'ooo'), 0),
    (('', None), 0),
    ((None, 'ooo'), 0),
    ((None, None), 0),
    ((5, None), 0),
    ((5, 5), 1),
    ((5.5, 5), 2),
    ((5.5, '5.'), 1),
    ((65.5, '5.'), 1),
    ((654.5, '5.'), 0),
    (('', ''), 1),
    (('1', ''), 2),
    ((1.4, ''), 4),
])
def test_count_substr(case, expected):
    assert _.count_substr(*case) == expected


@parametrize('case,expected', [
    ('\xC0\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xCA\xCB\xCC\xCD\xCE\xCF'
     '\xD0\xD1\xD2\xD3\xD4\xD5\xD6\xD7\xD8\xD9\xDA\xDB\xDC\xDD\xDE\xDF'
     '\xE0\xE1\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xEA\xEB\xEC\xED\xEE\xEF'
     '\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7\xF8\xF9\xFA\xFB\xFC\xFD\xFE\xFF',
     'AAAAAAAeCEEEEIIII'
     'DNOOOOO OUUUUYThss'
     'aaaaaaaeceeeeiiii'
     'dnooooo ouuuuythy'),
    ('abcABC', 'abcABC'),
    ('', ''),
    (None, ''),
])
def test_deburr(case, expected):
    assert _.deburr(case) == expected


@parametrize('case,expected', [
    ('Foo', 'foo'),
    ('Foo bar', 'foo bar'),
    ('Foo Bar', 'foo Bar'),
    ('FOO BAR', 'fOO BAR'),
    (None, ''),
])
def test_decapitalize(case, expected):
    assert _.decapitalize(case) == expected


@parametrize('case,expected', [
    (('abc', 'c'), True),
    (('abc', 'b'), False),
    (('abc', None), True),
    (('', 'b'), False),
    (('', None), True),
    ((None, 'b'), False),
    ((None, None), True),
    ((6.34, 4), True),
    ((6.34, 3), False),
    (('abc', 'c', 3), True),
    (('abc', 'c', 2), False),
    (('abc', 'b', 2), True),
    (('abc', 'b', 1), False),
    ((6.34, 'b', 1), False),
])
def test_ends_with(case, expected):
    assert _.ends_with(*case) == expected


@parametrize('case,expected', [
    ('abc<> &"\'`efg', 'abc&lt;&gt; &amp;&quot;&#39;&#96;efg'),
    ('abc', 'abc'),
    ('', ''),
    (None, ''),
])
def test_escape(case, expected):
    assert _.escape(case) == expected


@parametrize('case,expected', [
    ('[pydash](http://pydash.readthedocs.org/)',
     '\[pydash\]\(http\:\/\/pydash\.readthedocs\.org\/\)'),
    ('abc', 'abc'),
    ('', ''),
    (None, ''),
])
def test_escape_reg_exp(case, expected):
    assert _.escape_reg_exp(case) == expected


@parametrize('text,prefix,expected', [
    ('Hello world!', 'Hello', 'Hello world!'),
    (' world!', 'Hello', 'Hello world!'),
    ('', 'Hello', 'Hello'),
    ('', '', ''),
    ('1', '', '1'),
    ('1', '1', '1'),
    (5, 6, '65'),
    (None, 6, '6'),
    (None, None, ''),
])
def test_ensure_starts_with(text, prefix, expected):
    assert _.ensure_starts_with(text, prefix) == expected


@parametrize('text,suffix,expected', [
    ('Hello world!', 'world!', 'Hello world!'),
    ('Hello ', 'world!', 'Hello world!'),
    ('', 'Hello', 'Hello'),
    ('', '', ''),
    ('1', '', '1'),
    ('1', '1', '1'),
    (5, 6, '56'),
    (None, 6, '6'),
    (None, None, ''),
])
def test_ensure_ends_with(text, suffix, expected):
    assert _.ensure_ends_with(text, suffix) == expected


@parametrize('case,expected', [
    (('foobar', 'oo'), True),
    (('foobar', 'x'), False),
    (('foobar', 'f'), True),
    (('foobar', 'r'), True),
    (('foobar', ''), True),
    (('foobar', None), True),
    (('', ''), True),
    (('', None), True),
    ((None, None), True),
    ((56, 6), True),
    ((56, 7), False),
    ((5.6, '.'), True),
])
def test_has_substr(case, expected):
    assert _.has_substr(*case) == expected


@parametrize('case,expected', [
    ('  capitalize dash-CamelCase_underscore trim_id  ',
     'Capitalize dash camel case underscore trim'),
    ('foo_bar_id', 'Foo bar'),
    ('FooBar', 'Foo bar'),
    (5, '5'),
    ('', ''),
    (None, ''),
])
def test_human_case(case, expected):
    assert _.human_case(case) == expected


@parametrize('case,expected', [
    (('foobar', 0, 'xx'), 'xxfoobar'),
    (('foobar', 1, 'xx'), 'fxxoobar'),
    (('foobar', 4, 'xx'), 'foobxxar'),
    (('foobar', 6, 'xx'), 'foobarxx'),
    (('foobar', 7, 'xx'), 'foobarxx'),
    (('f', 7, 'xx'), 'fxx'),
    (('', 7, 'xx'), 'xx'),
    (('', 7, ''), ''),
    (('', 7, None), ''),
    ((None, 7, None), ''),
    ((None, 0, None), ''),
])
def test_insert_substr(case, expected):
    assert _.insert_substr(*case) == expected


@parametrize('case,expected', [
    (((1, 2, 3), '.'), '1.2.3'),
    ((('one', 'two', 'three'), '-.-'), 'one-.-two-.-three'),
    ((('s', 't', 'r', 'i', 'n', 'g'), ''), 'string'),
    ((('s', 't', 'r', 'i', 'n', 'g'),), 'string'),
    ((('s', 't', 'r', 'i', 'n', 'g'), None), 'string'),
    ((('string1', 'string2'), ','), 'string1,string2'),
    ((('string1', 'string2'), 5.6), 'string15.6string2'),
    (((None, 'string2'), 5.6), '5.6string2'),
    (((7, ), 5.6), '7'),
    (((None, ), 5.6), ''),
    (((None, None), 5.6), '5.6'),
    (((None, None, None), 5.6), '5.65.6'),
    (((None, None, None), None), ''),
    ((None, None), ''),
])
def test_join(case, expected):
    assert _.join(*case) == expected


@parametrize('case,expected', [
    ('foo  bar baz', 'foo-bar-baz'),
    ('foo__bar_baz', 'foo-bar-baz'),
    ('foo-_bar-_-baz', 'foo-bar-baz'),
    ('foo!bar,baz', 'foo-bar-baz'),
    ('--foo.bar;baz', 'foo-bar-baz'),
    ('Foo Bar', 'foo-bar'),
    ('fooBar', 'foo-bar'),
    (None, ''),
    (5, '5'),
    (5.6, '5-6'),
    (-5.6, '5-6'),
])
def test_kebab_case(case, expected):
    assert _.kebab_case(case) == expected


@parametrize('case,expected', [
    ('foo\nbar', ['foo', 'bar']),
    ('foo\rbar', ['foo', 'bar']),
    ('foo\r\nbar', ['foo', 'bar']),
    ('foo\n', ['foo']),
    ('\nfoo', ['', 'foo']),
    ('', []),
    (None, []),
])
def test_lines(case, expected):
    assert _.lines(case) == expected


@parametrize('case,expected', [
    ('fooBar', 'foo bar'),
    ('--foo-Bar--', 'foo bar'),
    ('*Foo*B_a*R', 'foo b a r'),
    ('*Foo10_B*Ar', 'foo 10 b ar'),
    ('/?*Foo10/;"B*Ar', 'foo 10 b ar'),
    ('/?F@O__o10456?>.B?>";Ar', 'f o o 10456 b ar'),
    ('FoO Bar', 'fo o bar'),
    ('F\n\soO Bar', 'f so o bar'),
    ('Foo1054665Bar', 'foo 1054665 bar')
])
def test_lower_case(case, expected):
    assert _.lower_case(case) == expected


@parametrize('case,expected', [
    ('Foobar', 'foobar'),
    ('Foo Bar', 'foo Bar'),
    ('1foobar', '1foobar'),
    (';foobar', ';foobar')
])
def test_lower_first(case, expected):
    assert _.lower_first(case) == expected


@parametrize('case,expected', [
    ((1234,), '1,234'),
    ((1234567890,), '1,234,567,890'),
    ((1234, 2), '1,234.00'),
    ((1234, 1), '1,234.0'),
    ((1234, 2, ',', '.'), '1.234,00'),
    (('1234',), ''),
])
def test_number_format(case, expected):
    assert _.number_format(*case) == expected


@parametrize('case,expected', [
    (('abc', 5, '12'), '1abc1'),
    (('abc', 8), '  abc   '),
    (('abc', 8, '_-'), '_-abc_-_'),
    (('abc', 3), 'abc'),
    (('', 3), '   '),
    ((' ', 3), '   '),
    ((None, 3), '   '),
])
def test_pad(case, expected):
    assert _.pad(*case) == expected


@parametrize('case,expected', [
    (('aaaaa', 3), 'aaaaa'),
    (('aaaaa', 6), ' aaaaa'),
    (('aaaaa', 10), '     aaaaa'),
    (('aaaaa', 6, 'b'), 'baaaaa'),
    (('aaaaa', 6, 'bc'), 'caaaaa'),
    (('aaaaa', 9, 'bc'), 'bcbcaaaaa'),
    (('a', 9, '12'), '12121212a'),
    (('a', 8, '12'), '2121212a'),
    (('', 8, '12'), '12121212'),
    ((None, 8, '12'), '12121212'),
])
def test_pad_start(case, expected):
    assert _.pad_start(*case) == expected


@parametrize('case,expected', [
    (('aaaaa', 3), 'aaaaa'),
    (('aaaaa', 6), 'aaaaa '),
    (('aaaaa', 10), 'aaaaa     '),
    (('aaaaa', 6, 'b'), 'aaaaab'),
    (('aaaaa', 6, 'bc'), 'aaaaab'),
    (('aaaaa', 9, 'bc'), 'aaaaabcbc'),
    (('a', 9, '12'), 'a12121212'),
    (('a', 8, '12'), 'a1212121'),
    (('', 8, '12'), '12121212'),
    ((None, 8, '12'), '12121212'),
])
def test_pad_end(case, expected):
    assert _.pad_end(*case) == expected


@parametrize('case,expected', [
    ('foo bar baz', 'FooBarBaz'),
    ('foo  bar baz', 'FooBarBaz'),
    ('foo__bar_baz', 'FooBarBaz'),
    ('foo-_bar-_-baz', 'FooBarBaz'),
    ('foo!bar,baz', 'FooBarBaz'),
    ('--foo.bar;baz', 'FooBarBaz'),
    ('', ''),
    (None, ''),
])
def test_pascal_case(case, expected):
    assert _.pascal_case(case) == expected


@parametrize('case,expected', [
    ('b', 'a'),
    ('B', 'A'),
])
def test_predecessor(case, expected):
    assert _.predecessor(case) == expected


@parametrize('case,expected', [
    (('Hello, world',), '...'),
    (('Hello, world', 5), 'Hello...'),
    (('Hello, world', 8), 'Hello...'),
    (('Hello, world', 5, ' (read a lot more)'), 'Hello, world'),
    (('Hello, cruel world', 15), 'Hello, cruel...'),
    (('Hello', 10), 'Hello'),
    (('', 10), ''),
    (('',), ''),
    ((None,), ''),
])
def test_prune(case, expected):
    assert _.prune(*case) == expected


@parametrize('case,expected', [
    (('hello world!',), '"hello world!"'),
    (('',), '""'),
    (('', None), ''),
    ((None,), '""'),
    ((None, None), ''),
    ((5,), '"5"'),
    ((5, None), '5'),
    ((-89,), '"-89"'),
    (('hello world!', '*'), '*hello world!*'),
    (('hello world!', '**'), '**hello world!**'),
    (('', '**'), '****'),
    (('hello world!', ''), 'hello world!'),
    ((28, '**'), '**28**'),
    ((28, ''), '28'),
    ((2, 8), '828'),
    ((-2, 8), '8-28'),
    ((-2, -8), '-8-2-8'),
    ((-8.5, 0), '0-8.50'),
])
def test_quote(case, expected):
    assert _.quote(*case) == expected


@parametrize('case,expected', [
    (('Hello World', '/[A-Z]/'), ['H']),
    (('Hello World', '/[A-Z]/g'), ['H', 'W']),
    (('hello world', '/[A-Z]/i'), ['h']),
    (('hello world', '/[A-Z]/gi'),
     ['h', 'e', 'l', 'l', 'o', 'w', 'o', 'r', 'l', 'd']),
    (('12345', '/[A-Z]/'), [])
])
def test_reg_exp_js_match(case, expected):
    assert _.reg_exp_js_match(*case) == expected


@parametrize('case,expected', [
    (('Hello World', '/[A-Z]/', '!'), '!ello World'),
    (('Hello World', '/[A-Z]/g', '!'), '!ello !orld'),
    (('hello world', '/[A-Z]/i', '!'), '!ello world'),
    (('hello world', '/[A-Z]/gi', '!'), '!!!!! !!!!!'),
    (('hello world', '/[A-Z]/gi', ''), ' '),
    (('hello world', '/[A-Z]/gi', None), ' '),
])
def test_reg_exp_js_replace(case, expected):
    assert _.reg_exp_js_replace(*case) == expected


@parametrize('case,expected', [
    (('foo', 'o', 'a'), 'faa'),
    (('foo', 'o', 'a', False, 1), 'fao'),
    (('fOO', 'o', 'a'), 'fOO'),
    (('fOO', 'o', 'a', True), 'faa'),
    (('', '', ''), ''),
    (('foo', 'o', ''), 'f'),
    (('foo', 'x', 'y'), 'foo'),
    (('foo', '^o', 'a'), 'foo'),
    (('ooo', '^o', 'f'), 'foo'),
    (('foo', 'o$', 'a'), 'foa'),
    (('foo', '', 'a'), 'afaoaoa'),
    (('foo', '', ''), 'foo'),
    (('foo', '', None), 'foo'),
    (('foo', None, None), 'foo'),
    (('foo', None, ''), 'foo'),
    (('foo', None, 'a'), 'foo'),
    ((54.7, 5, 6), '64.7'),
])
def test_reg_exp_replace(case, expected):
    assert _.reg_exp_replace(*case) == expected


@parametrize('case,expected', [
    (('foo',), ''),
    (('foo', 0), ''),
    (('foo', 1), 'foo'),
    (('foo', 3), 'foofoofoo'),
    (('', 3), ''),
    (('', 0), ''),
    ((None, 0), ''),
    ((None, 1), ''),
])
def test_repeat(case, expected):
    assert _.repeat(*case) == expected


@parametrize('case,expected', [
    (('foo', 'o', 'a'), 'faa'),
    (('foo', 'o', 'a', False, 1), 'fao'),
    (('fOO', 'o', 'a'), 'fOO'),
    (('fOO', 'o', 'a', True), 'faa'),
    (('', '', ''), ''),
    (('foo', 'o', ''), 'f'),
    (('foo', 'x', 'y'), 'foo'),
    (('foo', '', ''), 'foo'),
    (('foo', '', None), 'foo'),
    (('foo', None, None), 'foo'),
    (('foo', None, ''), 'foo'),
    (('foo', None, 'a'), 'foo'),
    ((54.7, 5, 6), '64.7'),
])
def test_replace(case, expected):
    assert _.replace(*case) == expected


@parametrize('case,expected', [
    (('foo', 'o', 'a'), 'foa'),
    (('foo', 'f', 'a'), 'foo'),
])
def test_replace_end(case, expected):
    assert _.replace_end(*case) == expected


@parametrize('case,expected', [
    (('foo', 'o', 'a'), 'foo'),
    (('foo', 'f', 'a'), 'aoo'),
])
def test_replace_start(case, expected):
    assert _.replace_start(*case) == expected


@parametrize('case,expected', [
    (('foo  bar baz', '-'), 'foo-bar-baz'),
    (('foo__bar_baz', '-'), 'foo-bar-baz'),
    (('foo-_bar-_-baz', '-'), 'foo-bar-baz'),
    (('foo!bar,baz', '-'), 'foo-bar-baz'),
    (('--foo.bar;baz', '-'), 'foo-bar-baz'),
    (('Foo Bar', '-'), 'foo-bar'),
    (('foo  bar baz', '_'), 'foo_bar_baz'),
    (('foo__bar_baz', '_'), 'foo_bar_baz'),
    (('foo-_bar-_-baz', '_'), 'foo_bar_baz'),
    (('foo!bar,baz', '_'), 'foo_bar_baz'),
    (('--foo.bar;baz', '_'), 'foo_bar_baz'),
    (('Foo Bar', '_'), 'foo_bar'),
    (('', '_'), ''),
    ((None, '_'), ''),
])
def test_separator_case(case, expected):
    assert _.separator_case(*case) == expected


@parametrize('case,expected', [
    (([],), ''),
    ((tuple(),), ''),
    (((None,),), ''),
    (((None, None),), ''),
    ((('', None),), ''),
    (((None, ''),), ''),
    (((None, 5),), '5'),
    (((7.88, None),), '7.88'),
    ((['', ''],), ''),
    ((['foo'],), 'foo'),
    ((['foo', 'bar'],), 'foo and bar'),
    ((['foo', 'bar', 'baz'],), 'foo, bar and baz'),
    ((['foo', 'bar', 'baz', 'qux'], ', ', ' or '), 'foo, bar, baz or qux'),
    ((['foo', 'bar', 'baz', 'qux'], ';', ' or '), 'foo;bar;baz or qux'),
    ((['foo', 'bar', 'baz', 'qux'], 0.6, ' or '), 'foo0.6bar0.6baz or qux'),
    ((['foo', 'bar', 'baz', 'qux'], 0.6, 1), 'foo0.6bar0.6baz1qux'),
    ((['foo', 'bar', 'baz', 'qux'], 0.6, None), 'foo0.6bar0.6bazqux'),
    ((['foo', 23, 'baz', 'qux'], None, None), 'foo23bazqux'),
])
def test_series_phrase(case, expected):
    assert _.series_phrase(*case) == expected


@parametrize('case,expected', [
    (([],), ''),
    ((tuple(),), ''),
    (((None,),), ''),
    (((None, None),), ''),
    ((('', None),), ''),
    (((None, ''),), ''),
    (((None, 5),), '5'),
    (((7.88, None),), '7.88'),
    ((['', ''],), ''),
    ((['foo'],), 'foo'),
    ((['foo', 'bar'],), 'foo and bar'),
    ((['foo', 'bar', 'baz'],), 'foo, bar, and baz'),
    ((['foo', 'bar', 'baz', 'qux'], ', ', ' or '), 'foo, bar, baz, or qux'),
    ((['foo', 'bar', 'baz', 'qux'], ';', ' or '), 'foo;bar;baz; or qux'),
])
def test_series_phrase_serial(case, expected):
    assert _.series_phrase_serial(*case) == expected


@parametrize('case,expected', [
    ('Foo Bar', 'foo-bar'),
    (' foo bar ', 'foo-bar'),
    (u'Un éléphant à l\'orée du bois', 'un-elephant-a-l-oree-du-bois'),
    ('', ''),
    (5, '5'),
    (None, ''),
])
def test_slugify(case, expected):
    assert _.slugify(case) == expected


@parametrize('case,expected', [
    ('foo  bar baz', 'foo_bar_baz'),
    ('foo__bar_baz', 'foo_bar_baz'),
    ('foo-_bar-_-baz', 'foo_bar_baz'),
    ('foo!bar,baz', 'foo_bar_baz'),
    ('--foo.bar;baz', 'foo_bar_baz'),
    ('FooBar', 'foo_bar'),
    ('fooBar', 'foo_bar'),
    ('', ''),
    (None, ''),
    (5, '5'),
])
def test_snake_case(case, expected):
    assert _.snake_case(case) == expected


@parametrize('case,expected', [
    (('string1 string2',), ['string1', 'string2']),
    (('string', ''), ['s', 't', 'r', 'i', 'n', 'g']),
    (('string', None), ['s', 't', 'r', 'i', 'n', 'g']),
    (('', ''), []),
    (('', None), []),
    ((None, None), []),
    (('string1,string2', ','), ['string1', 'string2']),
])
def test_split(case, expected):
    assert _.split(*case) == expected


@parametrize('case,expected', [
    ('foo bar baz', 'Foo Bar Baz'),
    ('foo-bar-baz', 'Foo Bar Baz'),
    ("Foo!#Bar's", 'Foo Bar S'),
    ('fooBar', 'Foo Bar'),
    ('___FOO-BAR___', 'FOO BAR'),
    ('XMLHttp', 'XML Http'),
    ("", ""),
    (None, ""),
    (5, "5"),
])
def test_start_case(case, expected):
    assert _.start_case(case) == expected


@parametrize('case,expected', [
    (('abc', 'a'), True),
    (('abc', 'b'), False),
    (('abc', 'a', 0), True),
    (('abc', 'a', 1), False),
    (('abc', 'b', 1), True),
    (('abc', 'b', 2), False),
    ((5.78, 5), True),
    (("5.78", 5), True),
    ((5.78, '5'), True),
    ((5.78, ''), True),
    ((5.78, None), True),
    ((None, None), True),
    (('', None), True),
    ((' ', None), True),
])
def test_starts_with(case, expected):
    assert _.starts_with(*case) == expected


@parametrize('case,expected', [
    ('a <a href="#">link</a>', 'a link'),
    ('a <a href="#">link</a><script>alert("hello world!")</script>',
     'a linkalert("hello world!")'),
    ('', ''),
    (None, ''),
])
def test_strip_tags(case, expected):
    assert _.strip_tags(case) == expected


@parametrize('case,expected', [
    (('This_is_a_test_string', '_'), 'This'),
    (('This_is_a_test_string', ''), 'This_is_a_test_string'),
    (('This_is_a_test_string', ' '), 'This_is_a_test_string'),
    (('This_is_a_test_string', None), 'This_is_a_test_string'),
    ((None, None), ''),
    ((None, '4'), ''),
    ((None, ''), ''),
])
def test_substr_left(case, expected):
    assert _.substr_left(*case) == expected


@parametrize('case,expected', [
    (('This_is_a_test_string', '_'), 'This_is_a_test'),
    (('This_is_a_test_string', ''), 'This_is_a_test_string'),
    (('This_is_a_test_string', ' '), 'This_is_a_test_string'),
    (('This_is_a_test_string', None), 'This_is_a_test_string'),
    ((None, None), ''),
    ((None, '4'), ''),
    ((None, ''), ''),
])
def test_substr_left_end(case, expected):
    assert _.substr_left_end(*case) == expected


@parametrize('case,expected', [
    (('This_is_a_test_string', '_'), 'is_a_test_string'),
    (('This_is_a_test_string', ''), 'This_is_a_test_string'),
    (('This_is_a_test_string', ' '), 'This_is_a_test_string'),
    (('This_is_a_test_string', None), 'This_is_a_test_string'),
    ((None, None), ''),
    ((None, '4'), ''),
    ((None, ''), ''),
])
def test_substr_right(case, expected):
    assert _.substr_right(*case) == expected


@parametrize('case,expected', [
    (('This_is_a_test_string', '_'), 'string'),
    (('This_is_a_test_string', ''), 'This_is_a_test_string'),
    (('This_is_a_test_string', ' '), 'This_is_a_test_string'),
    (('This_is_a_test_string', None), 'This_is_a_test_string'),
    ((None, None), ''),
    ((None, '4'), ''),
    ((None, ''), ''),
])
def test_substr_right_end(case, expected):
    assert _.substr_right_end(*case) == expected


@parametrize('case,expected', [
    ('a', 'b'),
    ('A', 'B'),
])
def test_successor(case, expected):
    assert _.successor(case) == expected


@parametrize('source,wrapper,expected', [
    ('hello world!', '*', '*hello world!*'),
    ('hello world!', '**', '**hello world!**'),
    ('', '**', '****'),
    ('hello world!', '', 'hello world!'),
    (5, '1', '151'),
    (5, 1, '151'),
    ('5', 1, '151'),
    ('5', 12, '12512'),
    (5, '', '5'),
    ('', 5, '55'),
    ('', None, ''),
    (None, None, ''),
    (5, None, '5'),
])
def test_surround(source, wrapper, expected):
    assert _.surround(source, wrapper) == expected


@parametrize('case,expected', [
    ('fOoBaR', 'FoObAr'),
    ('', ''),
    (None, ''),
])
def test_swap_case(case, expected):
    assert _.swap_case(case) == expected


@parametrize('case,expected', [
    ('foo bar baz', 'Foo Bar Baz'),
    ("they're bill's friends from the UK",
     "They're Bill's Friends From The Uk"),
    ("", ""),
    (None, ""),
    (5, "5"),
])
def test_title_case(case, expected):
    assert _.title_case(case) == expected


@parametrize('case,expected', [
    ('--Foo-Bar--', '--foo-bar--'),
    ('fooBar', 'foobar'),
    ('__FOO_BAR__', '__foo_bar__')
])
def test_to_lower(case, expected):
    assert _.to_lower(case) == expected


@parametrize('case,expected', [
    ('--Foo-Bar--', '--FOO-BAR--'),
    ('fooBar', 'FOOBAR'),
    ('__FOO_BAR__', '__FOO_BAR__')
])
def test_to_upper(case, expected):
    assert _.to_upper(case) == expected


@parametrize('case,expected', [
    (('  fred  ',), 'fred'),
    (('-_-fred-_-', '_-'), 'fred'),
    (('-_-fred-_-', None), '-_-fred-_-'),
    ((None,), ''),
    ((None, None), ''),
])
def test_trim(case, expected):
    assert _.trim(*case) == expected


@parametrize('case,expected', [
    (('  fred  ',), 'fred  '),
    (('-_-fred-_-', '_-'), 'fred-_-'),
    (('-_-fred-_-', None), '-_-fred-_-'),
    ((None,), ''),
    ((None, None), ''),
])
def test_trim_start(case, expected):
    assert _.trim_start(*case) == expected


@parametrize('case,expected', [
    (('  fred  ',), '  fred'),
    (('-_-fred-_-', '_-'), '-_-fred'),
    (('-_-fred-_-', None), '-_-fred-_-'),
    ((None,), ''),
    ((None, None), ''),
])
def test_trim_end(case, expected):
    assert _.trim_end(*case) == expected


@parametrize('case,expected', [
    (('hi-diddly-ho there, neighborino',), 'hi-diddly-ho there, neighbo...'),
    (('hi-diddly-ho there, neighborino', 24), 'hi-diddly-ho there, n...'),
    (('hi-diddly-ho there, neighborino', 24, '...', ' '),
     'hi-diddly-ho there,...'),
    (('hi-diddly-ho there, neighborino', 24, '...', re.compile(',? +')),
     'hi-diddly-ho there...'),
    (('hi-diddly-ho there, neighborino', 30, ' [...]'),
     'hi-diddly-ho there, neig [...]'),
    (('123456789', 9), '123456789'),
    (('123456789', 8), '12345...'),
    (('x',), 'x'),
    ((' ',), ' '),
    (('',), ''),
    ((None,), ''),
])
def test_truncate(case, expected):
    assert _.truncate(*case) == expected


@parametrize('case,expected', [
    ('abc&lt;&gt; &amp;&quot;&#39;&#96;efg', 'abc<> &"\'`efg'),
    ('', ''),
    (' ', ' '),
    (None, ''),
])
def test_unescape(case, expected):
    assert _.unescape(case) == expected


@parametrize('case,expected', [
    (('"foo"',), 'foo'),
    (("'foo'", "'"), 'foo'),
    (('"foo',), '"foo'),
    (('foo"',), 'foo"'),
    ((' ',), ' '),
    (('',), ''),
    ((None,), ''),
])
def test_unquote(case, expected):
    assert _.unquote(*case) == expected


@parametrize('case,expected', [
    ('fooBar', 'FOO BAR'),
    ('--foo-Bar--', 'FOO BAR'),
    ('*Foo*B_a*R', 'FOO B A R'),
    ('*Foo10_B*Ar', 'FOO 10 B AR'),
    ('/?*Foo10/;"B*Ar', 'FOO 10 B AR'),
    ('/?F@O__o10456?>.B?>";Ar', 'F O O 10456 B AR'),
    ('FoO Bar', 'FO O BAR'),
    ('F\n\soO Bar', 'F SO O BAR'),
    ('Foo1054665Bar', 'FOO 1054665 BAR')
])
def test_upper_case(case, expected):
    assert _.upper_case(case) == expected


@parametrize('case,expected', [
    ('foobar', 'Foobar'),
    ('Foobar', 'Foobar'),
    ('1foobar', '1foobar'),
    (';foobar', ';foobar')
])
def test_upper_first(case, expected):
    assert _.upper_first(case) == expected


@parametrize('case,expected', [
    ({'args': [''], 'kargs': {}}, ''),
    ({'args': ['/'], 'kargs': {}}, '/'),
    ({'args': ['http://github.com'], 'kargs': {}}, 'http://github.com'),
    ({'args': ['http://github.com:80'], 'kargs': {}}, 'http://github.com:80'),
    ({'args': ['http://github.com:80', 'pydash', 'issues/'], 'kargs': {}},
     'http://github.com:80/pydash/issues/'),
    ({'args': ['/foo', 'bar', '/baz', '/qux/'], 'kargs': {}},
     '/foo/bar/baz/qux/'),
    ({'args': ['/foo/bar'], 'kargs': {'a': 1, 'b': 'two'}},
     '/foo/bar?a=1&b=two'),
    ({'args': ['/foo/bar?x=5'], 'kargs': {'a': 1, 'b': 'two'}},
     '/foo/bar?x=5&a=1&b=two'),
    ({'args': ['/foo/bar?x=5', 'baz?z=3'], 'kargs': {'a': 1, 'b': 'two'}},
     '/foo/bar/baz?x=5&a=1&b=two&z=3'),
    ({'args': ['/foo/bar?x=5', 'baz?z=3'], 'kargs': {'a': [1, 2], 'b': 'two'}},
     '/foo/bar/baz?x=5&a=1&a=2&b=two&z=3'),
    ({'args': ['/foo#bar', 'baz'], 'kargs': {'a': [1, 2], 'b': 'two'}},
     '/foo?a=1&a=2&b=two#bar/baz'),
    ({'args': ['/foo', 'baz#bar'], 'kargs': {'a': [1, 2], 'b': 'two'}},
     '/foo/baz?a=1&a=2&b=two#bar'),
])
def test_url(case, expected):
    result = _.url(*case['args'], **case['kargs'])

    r_scheme, r_netloc, r_path, r_query, r_fragment = urlsplit(result)
    e_scheme, e_netloc, e_path, e_query, e_fragment = urlsplit(expected)

    assert r_scheme == e_scheme
    assert r_netloc == e_netloc
    assert r_path == e_path
    assert set(parse_qsl(r_query)) == set(parse_qsl(e_query))
    assert r_fragment == e_fragment


@parametrize('case,expected', [
    ('hello world!', ['hello', 'world']),
    ('hello_world', ['hello', 'world']),
    ('hello!@#$%^&*()_+{}|:"<>?-=[]\;\,.\'/world', ['hello', 'world']),
    ('hello 12345 world', ['hello', '12345', 'world']),
    ('enable 24h format', ['enable', '24', 'h', 'format']),
    ('tooLegit2Quit', ['too', 'Legit', '2', 'Quit']),
    ('xhr2Request', ['xhr', '2', 'Request']),
    (' ', []),
    ('', []),
    (None, []),
])
def test_words(case, expected):
    assert _.words(case) == expected


@parametrize('case,expected', [
    ('enable 24h format', 'enable24HFormat'),
])
def test_word_cycle(case, expected):
    actual = (_.chain(case)
              .camel_case()
              .kebab_case()
              .snake_case()
              .start_case()
              .camel_case()
              .value())

    assert actual == expected



# -*- coding: utf-8 -*-

import pytest

import pydash as _
from pydash._compat import PY2


if PY2:
    typed_function = None
else:
    # Hack around the fact that we can't define this kind of function in
    # Python 2 so have to rely on conditionally creating it using exec()
    # to avoid syntax errors while still having this test case covered for
    # Python 3.
    exec('def typed_function(row: int, index: int, col: list): return row + 1')


@pytest.mark.skipif(PY2, reason='test requires Python 3 annotations')
def test_annotated_iteratee():
    assert _.map_([1, 2], typed_function) == [2, 3]



# -*- coding: utf-8 -*-
"""Exception classes.

.. versionadded:: 1.0.0
"""


class InvalidMethod(Exception):
    """Raised when an invalid pydash method is invoked through
    :func:`pydash.chaining.chain`.

    .. versionadded:: 1.0.0
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def parse_requirements(filename):
    return [line.strip()
            for line in read(filename).strip().split('\n')
            if line.strip()]

pkg = {}
exec(read('pydash/__pkg__.py'), pkg)

readme = read('README.rst')
changelog = read('CHANGELOG.rst')
requirements = parse_requirements('requirements.txt')


setup(
    name=pkg['__package_name__'],
    version=pkg['__version__'],
    url=pkg['__url__'],
    license=pkg['__license__'],
    author=pkg['__author__'],
    author_email=pkg['__email__'],
    description=pkg['__description__'],
    long_description=readme + '\n\n' + changelog,
    packages=find_packages(exclude=['tests', 'tasks']),
    install_requires=requirements,
    keywords='utility functional lodash underscore',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)



# -*- coding: utf-8 -*-

import datetime
import decimal
import operator
import re

import pydash as _

from . import fixtures
from .fixtures import parametrize


@parametrize('value,other,expected', [
    ('a', 'a', True),
    (None, None, True),
    (None, '', False),
    (1, str(1), False)
])
def test_eq(value, other, expected):
    assert _.eq(value, other) == expected


@parametrize('case,expected', [
    ((2, 1), True),
    ((2.5, 2.3), True),
    ((1, 2), False),
    ((2.3, 2.5), False),
    ((1, 1), False),
])
def test_gt(case, expected):
    assert _.gt(*case) == expected


@parametrize('case,expected', [
    ((2, 1), True),
    ((2.5, 2.3), True),
    ((1, 2), False),
    ((2.3, 2.5), False),
    ((1, 1), True),
])
def test_gte(case, expected):
    assert _.gte(*case) == expected


@parametrize('case,expected', [
    ((2, 1), False),
    ((2.5, 2.3), False),
    ((1, 2), True),
    ((2.3, 2.5), True),
    ((1, 1), False),
])
def test_lt(case, expected):
    assert _.lt(*case) == expected


@parametrize('case,expected', [
    ((2, 1), False),
    ((2.5, 2.3), False),
    ((1, 2), True),
    ((2.3, 2.5), True),
    ((1, 1), True),
])
def test_lte(case, expected):
    assert _.lte(*case) == expected


@parametrize('case,expected', [
    ((3, 2, 4), True),
    ((4, 8), True),
    ((4, 2), False),
    ((2, 2), False),
    ((1.2, 2), True),
    ((5.2, 4), False),
    (('', 5), False),
    ((2, ''), False),
    ((-1, -2, ''), True),
])
def test_in_range(case, expected):
    assert _.in_range(*case) == expected


@parametrize('case,expected', [
    ([], True),
    ({}, True),
    ('', True),
    (0, False),
    (True, False),
])
def test_is_associative(case, expected):
    assert _.is_associative(case) == expected


@parametrize('case,expected', [
    ('', True),
    ('\n', True),
    (' ', True),
    ('a', False),
])
def test_is_blank(case, expected):
    assert _.is_blank(case) == expected


@parametrize('case,expected', [
    (True, True),
    (False, True),
    (0, False),
    ('', False),
])
def test_is_boolean(case, expected):
    assert _.is_boolean(case) == expected


@parametrize('case,expected', [
    (list, True),
    (dict, True),
    ([].append, True),
    ({}.update, True),
    (1, True),
    (True, True),
    (None, True),
    (datetime, False),
    (parametrize, False),
    ('a', False),
    (lambda: None, False),
])
def test_is_builtin(case, expected):
    assert _.is_builtin(case) == expected


@parametrize('case,expected', [
    (datetime.date.today(), True),
    (datetime.datetime.today(), True),
    ('2014-01-01', False),
    ('2014-01-01T00:00:00', False),
    (1, False)
])
def test_is_date(case, expected):
    assert _.is_date(case) == expected


@parametrize('case,expected', [
    ([3, 2, 1], True),
    ([6, 5, 5], True),
    (0, True),
    ([5, 4, 4, 3, 1], True),
    ([5, 4, 4, 5, 4, 3], False)
])
def test_is_decreasing(case, expected):
    assert _.is_decreasing(case) == expected


@parametrize('case,expected', [
    ({}, True),
    ([], False),
    (1, False),
    ('a', False),
    (iter([]), False),
    (iter({}), False)
])
def test_is_dict(case, expected):
    assert _.is_dict(case) == expected


@parametrize('case,expected', [
    (True, True),
    (0, True),
    (123.45, True),
    ('', True),
    ({}, True),
    ([], True),
    (False, True),
    (None, True),
    ({'a': 1}, False),
    ([1], False),
    ('Hello', False),
    (['Hello', 'World'], False),
])
def test_is_empty(case, expected):
    assert _.is_empty(case) == expected


@parametrize('case,expected', [
    ((1, 1), True),
    ((1, 2), False),
    (('1', '1'), True),
    (('1', '2'), False),
    (([1], {'a': 1}), False),
    (({'a': 1}, {'a': 1}), True),
    (({'a': 1}, {'b': 1}), False),
    (([1, 2, 3], [1, 2, 3]), True),
    (([1, 2, 3], [1, 2]), False),
    (([1, 2], [1, 2, 3]), False),
    (([1, 2, 3], [1, 2]), False),
    (([1, 2], [1, 2, 3]), False),
    ((['hello', 'goodbye'], ['hi', 'goodbye']), False)
])
def test_is_equal(case, expected):
    assert _.is_equal(*case) == expected


@parametrize('case,expected', [
    ((1, 1, None), True),
    ((1, 2, None), False),
    (('1', '1', None), True),
    (('1', '2', None), False),
    (([1], {'a': 1}, None), False),
    (([1], {'a': 1}, lambda a, b: True), True),
    (({'a': 1}, {'a': 1}, None), True),
    (({'a': 1}, {'b': 1}, lambda a, b: None if isinstance(a, dict) else True),
     False),
    (([1, 2, 3], [1, 2, 3], None), True),
    (([1, 2, 3], [1, 2], None), False),
    (([1, 2], [1, 2, 3], None), False),
    (([1, 2, 3], [1, 2], lambda a, b: None if isinstance(a, list) else True),
     False),
    (([1, 2], [1, 2, 3], lambda a, b: None if isinstance(a, list) else True),
     False),
    ((['hello', 'goodbye'], ['hi', 'goodbye'], fixtures.is_equal_iteratee0),
     True)
])
def test_is_equal_with(case, expected):
    assert _.is_equal_with(*case) == expected


@parametrize('case,expected', [
    (Exception(), True),
    ({}, False),
    ([], False)
])
def test_is_error(case, expected):
    assert _.is_error(case) == expected


@parametrize('case,expected', [
    (2, True),
    (16, True),
    (1, False),
    (3.0, False),
    (3.5, False),
    (None, False),
])
def test_is_even(case, expected):
    assert _.is_even(case) == expected


@parametrize('case,expected', [
    (1.0, True),
    (3.245, True),
    (1, False),
    (True, False),
    ('', False),
])
def test_is_float(case, expected):
    assert _.is_float(case) == expected


@parametrize('case,expected', [
    (lambda x: x + 1, True),
    ('Hello, world!', False),
])
def test_is_function(case, expected):
    assert _.is_function(case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], True),
    ([5, 5, 6], True),
    (0, True),
    ([1, 2, 3, 4, 4, 5], True),
    ([1, 2, 3, 4, 4, 3], False)
])
def test_is_increasing(case, expected):
    assert _.is_increasing(case) == expected


@parametrize('case,expected', [
    ([], True),
    ('', True),
    ((), True),
    ({}, False),
    (1, False),
    (None, False),
])
def test_is_indexed(case, expected):
    assert _.is_indexed(case) == expected


@parametrize('case,expected', [
    ((1, int), True),
    ((1.0, float), True),
    (([], (list, str)), True),
    (([], dict), False),
    ((True, float), False),
])
def test_is_instance_of(case, expected):
    assert _.is_instance_of(*case) == expected


@parametrize('case,expected', [
    (1, True),
    (2, True),
    (1.0, False),
    (True, False),
    (None, False),
    ([], False),
])
def test_is_integer(case, expected):
    assert _.is_integer(case) == expected


@parametrize('case,expected', [
    ([], True),
    ({}, True),
    ((), True),
    ('a', True),
    (5, False),
    (None, False),
])
def test_is_iterable(case, expected):
    assert _.is_iterable(case) == expected


@parametrize('case,expected', [
    ('{"one": 1, "two": {"three": "3"}, "four": [4]}', True),
    ({"one": 1, "two": {"three": "3"}, "four": [4]}, False),
    ('', False),
    (1, False),
    (True, False),
])
def test_is_json(case, expected):
    assert _.is_json(case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], True),
    ({}, False),
])
def test_is_list(case, expected):
    assert _.is_list(case) == expected


@parametrize('case,expected', [
    (({'name': 'fred', 'age': 40}, {'age': 40}), True),
    (({'name': 'fred', 'age': 40}, {'age': 40, 'active': True}), False),
    (([1, 2, 3], [1, 2]), True),
    (([1, 2, 3], [1, 2, 3, 4]), False),
    (({}, {}), True),
    (({'a': 1}, {}), True),
    (([], []), True),
    (([1], []), True),
])
def test_is_match(case, expected):
    assert _.is_match(*case) == expected


@parametrize('case,expected', [
    (([1, 2],
      [2, 4],
      lambda a, b: None if isinstance(a, list) else b == a + a),
     True),
    (([1, 2],
      [2, 4],
      lambda a, b, key: a == b == key),
     False),
    (([0, 1],
      [0, 1],
      lambda a, b, key: a == b == key),
     True),
])
def test_is_match_with(case, expected):
    assert _.is_match_with(*case) == expected


@parametrize('case,expected', [
    (([1, 2, 3], operator.le), True),
    (([3, 2, 1], operator.ge), True),
    (([1, 1, 2], operator.lt), False),
    (([3, 3, 2], operator.gt), False),
])
def test_is_monotone(case, expected):
    assert _.is_monotone(*case) == expected


@parametrize('case,expected', [
    (0, False),
    (123456789123456789123456789, False),
    (123.45, False),
    (decimal.Decimal(1), False),
    ('1', True)
])
def test_is_nan(case, expected):
    assert _.is_nan(case) == expected


@parametrize('case,expected', [
    (-1, True),
    (-1.25, True),
    (-0.1, True),
    (0, False),
    (1, False),
    (True, False),
    (False, False)
])
def test_is_negative(case, expected):
    assert _.is_negative(case) == expected


@parametrize('case,expected', [
    (None, True),
    (0, False),
])
def test_is_none(case, expected):
    assert _.is_none(case) == expected


@parametrize('case,expected', [
    (0, True),
    (123456789123456789123456789, True),
    (123.45, True),
    (decimal.Decimal(1), True),
    ('1', False),
    (True, False),
    (False, False),
])
def test_is_number(case, expected):
    assert _.is_number(case) == expected


@parametrize('case,expected', [
    ({}, True),
    ([], True),
    (1, False),
    ('a', False),
    (iter([]), False),
    (iter({}), False)
])
def test_is_object(case, expected):
    assert _.is_object(case) == expected


@parametrize('case,expected', [
    (1, True),
    (3.0, True),
    (3.5, True),
    (2, False),
    (16, False),
    (None, False),
])
def test_is_odd(case, expected):
    assert _.is_odd(case) == expected


@parametrize('case,expected', [
    (1, True),
    (1.25, True),
    (0.1, True),
    (-1, False),
    (0, False),
    (True, False),
    (False, False)
])
def test_is_positive(case, expected):
    assert _.is_positive(case) == expected


@parametrize('case,expected', [
    (re.compile(''), True),
    ('', False),
    ('Hello, world!', False),
    (1, False),
    ({}, False),
    ([], False),
    (None, False)
])
def test_is_reg_exp(case, expected):
    assert _.is_reg_exp(case) == expected


@parametrize('case,expected', [
    (set(), True),
    ([1, 2, 3], False)
])
def test_is_set(case, expected):
    assert _.is_set(case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], False),
    ([3, 2, 1], True),
    ([1, 1, 2], False),
    ([3, 3, 2], False),
])
def test_is_strictly_decreasing(case, expected):
    assert _.is_strictly_decreasing(case) == expected


@parametrize('case,expected', [
    ([1, 2, 3], True),
    ([3, 2, 1], False),
    ([1, 1, 2], False),
    ([3, 3, 2], False),
])
def test_is_strictly_increasing(case, expected):
    assert _.is_strictly_increasing(case) == expected


@parametrize('case,expected', [
    ('', True),
    ('Hello, world!', True),
    (1, False),
    ({}, False),
    ([], False),
    (None, False)
])
def test_is_string(case, expected):
    assert _.is_string(case) == expected


@parametrize('case,expected', [
    ((), True),
    ([], False),
    ({}, False),
])
def test_is_tuple(case, expected):
    assert _.is_tuple(case) == expected


@parametrize('case,expected', [
    (0, True),
    (0.0, False),
    ('', False),
    (True, False),
    (False, False),
])
def test_is_zero(case, expected):
    assert _.is_zero(case) == expected



# -*- coding: utf-8 -*-
"""Python port of Lo-Dash
"""

from .__pkg__ import (
    __description__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__
)


#
# Arrays
#

from .arrays import (
    chunk,
    compact,
    concat,
    difference,
    difference_by,
    difference_with,
    drop,
    drop_right,
    drop_right_while,
    drop_while,
    duplicates,
    fill,
    find_index,
    find_last_index,
    flatten,
    flatten_deep,
    flatten_depth,
    from_pairs,
    head,
    index_of,
    initial,
    intercalate,
    interleave,
    intersection,
    intersection_by,
    intersection_with,
    intersperse,
    last,
    last_index_of,
    mapcat,
    nth,
    pop,
    pull,
    pull_all,
    pull_all_by,
    pull_all_with,
    pull_at,
    push,
    remove,
    reverse,
    shift,
    slice_,
    sort,
    sorted_index,
    sorted_index_by,
    sorted_index_of,
    sorted_last_index,
    sorted_last_index_by,
    sorted_last_index_of,
    sorted_uniq,
    sorted_uniq_by,
    splice,
    split_at,
    tail,
    take,
    take_right,
    take_right_while,
    take_while,
    union,
    union_by,
    union_with,
    uniq,
    uniq_by,
    uniq_with,
    unshift,
    unzip,
    unzip_with,
    without,
    xor,
    xor_by,
    xor_with,
    zip_,
    zip_with,
    zip_object,
    zip_object_deep
)


#
# Chaining
#

from .chaining import (
    chain,
    tap,
    thru,
    _Dash
)


#
# Collections
#

from .collections import (
    at,
    count_by,
    every,
    filter_,
    find,
    find_last,
    flat_map,
    flat_map_deep,
    flat_map_depth,
    for_each,
    for_each_right,
    group_by,
    includes,
    invoke_map,
    key_by,
    map_,
    nest,
    order_by,
    partition,
    pluck,
    reduce_,
    reduce_right,
    reductions,
    reductions_right,
    reject,
    sample,
    sample_size,
    shuffle,
    size,
    some,
    sort_by
)


#
# Functions
#

from .functions import (
    after,
    ary,
    before,
    conjoin,
    curry,
    curry_right,
    debounce,
    delay,
    disjoin,
    flip,
    flow,
    flow_right,
    iterated,
    juxtapose,
    negate,
    once,
    over_args,
    partial,
    partial_right,
    rearg,
    spread,
    throttle,
    unary,
    wrap,
)


#
# Numerical
#

from .numerical import (
    add,
    ceil,
    clamp,
    divide,
    floor,
    max_,
    max_by,
    mean,
    mean_by,
    median,
    min_,
    min_by,
    moving_mean,
    multiply,
    power,
    round_,
    scale,
    slope,
    std_deviation,
    subtract,
    sum_,
    sum_by,
    transpose,
    variance,
    zscore,
)


#
# Objects
#

from .objects import (
    assign,
    assign_with,
    callables,
    clone,
    clone_deep,
    clone_deep_with,
    clone_with,
    defaults,
    defaults_deep,
    find_key,
    find_last_key,
    for_in,
    for_in_right,
    get,
    has,
    invert,
    invert_by,
    invoke,
    keys,
    map_keys,
    map_values,
    map_values_deep,
    merge,
    merge_with,
    omit,
    omit_by,
    parse_int,
    pick,
    pick_by,
    rename_keys,
    set_,
    set_with,
    to_boolean,
    to_dict,
    to_integer,
    to_list,
    to_number,
    to_pairs,
    to_string,
    transform,
    unset,
    update,
    update_with,
    values
)


#
# Predicates
#

from .predicates import (
    eq,
    gt,
    gte,
    lt,
    lte,
    in_range,
    is_associative,
    is_blank,
    is_boolean,
    is_builtin,
    is_date,
    is_decreasing,
    is_dict,
    is_empty,
    is_equal,
    is_equal_with,
    is_error,
    is_even,
    is_float,
    is_function,
    is_increasing,
    is_indexed,
    is_instance_of,
    is_integer,
    is_iterable,
    is_json,
    is_list,
    is_match,
    is_match_with,
    is_monotone,
    is_nan,
    is_negative,
    is_none,
    is_number,
    is_object,
    is_odd,
    is_positive,
    is_reg_exp,
    is_set,
    is_strictly_decreasing,
    is_strictly_increasing,
    is_string,
    is_tuple,
    is_zero,
)


#
# Strings
#

from .strings import (
    camel_case,
    capitalize,
    chars,
    chop,
    chop_right,
    clean,
    count_substr,
    deburr,
    decapitalize,
    ends_with,
    ensure_starts_with,
    ensure_ends_with,
    escape,
    escape_reg_exp,
    has_substr,
    human_case,
    insert_substr,
    join,
    kebab_case,
    lines,
    lower_case,
    lower_first,
    number_format,
    pad,
    pad_end,
    pad_start,
    pascal_case,
    predecessor,
    prune,
    quote,
    reg_exp_js_match,
    reg_exp_js_replace,
    reg_exp_replace,
    repeat,
    replace,
    replace_end,
    replace_start,
    separator_case,
    series_phrase,
    series_phrase_serial,
    slugify,
    snake_case,
    split,
    start_case,
    starts_with,
    strip_tags,
    substr_left,
    substr_left_end,
    substr_right,
    substr_right_end,
    successor,
    surround,
    swap_case,
    title_case,
    to_lower,
    to_upper,
    trim,
    trim_end,
    trim_start,
    truncate,
    unescape,
    unquote,
    upper_case,
    upper_first,
    url,
    words,
)

#
# Utilities
#

from .utilities import (
    attempt,
    cond,
    conforms,
    conforms_to,
    constant,
    default_to,
    iteratee,
    identity,
    matches,
    matches_property,
    memoize,
    method,
    method_of,
    noop,
    now,
    nth_arg,
    over,
    over_every,
    over_some,
    properties,
    property_,
    property_of,
    random,
    range_,
    range_right,
    result,
    retry,
    stub_list,
    stub_dict,
    stub_false,
    stub_string,
    stub_true,
    times,
    to_path,
    unique_id,
)


#
# Exceptions
#

from .exceptions import (
    InvalidMethod
)


#
# "_" Instance
#

py_ = _Dash()
_ = py_
